import numpy as np
import time as tm
import re
import warnings
from .. import Utils
from discretize import TreeMesh


def read_GOCAD_ts(tsfile):
    """

    Read GOCAD triangulated surface (*.ts) file
    INPUT:
    tsfile: Triangulated surface

    OUTPUT:
    vrts : Array of vertices in XYZ coordinates [n x 3]
    trgl : Array of index for triangles [m x 3]. The order of the vertices
            is important and describes the normal
            n = cross( (P2 - P1 ) , (P3 - P1) )

    Author: @fourndo


    .. note::

        Remove all attributes from the GoCAD surface before exporting it!

    """

    fid = open(tsfile)
    line = fid.readline()

    # Skip all the lines until the vertices
    VRTX, TRGL = [], []
    while "END" not in line:

        while "VRTX" not in line:
            line = fid.readline()
            if "END\n" in line:
                return VRTX, TRGL

        vrtx = []
        # Run down all the vertices and save in array
        while np.any(["VRTX" in line, "PVRTX" in line]):
            l_input = re.split(r"[\s*]", line)
            temp = np.array(l_input[2:5])
            vrtx.append(temp.astype(np.float))

            # Read next line
            line = fid.readline()

        VRTX += [np.asarray(vrtx)]

        # Skip lines to the triangles
        while "TRGL" not in line:
            line = fid.readline()

        # Run down the list of triangles
        trgl = []

        # Run down all the vertices and save in array
        while "TRGL" in line:
            l_input = re.split(r"[\s*]", line)
            temp = np.array(l_input[1:4])
            trgl.append(temp.astype(np.int))

            # Read next line
            line = fid.readline()

        TRGL += [np.asarray(trgl)]

    return VRTX, TRGL


def read_GOCAD_pl(plFile):
    """

    Read GOCAD polyline file (*.pl)
    INPUT:
    plFile: Polyline object

    OUTPUT:
    vrts : List Array of vertices in XYZ coordinates [n x 3]
    segs : List Array of index for segments [m x 3]. The order of the vertices
            is important and describes the normal
            n = cross( (P2 - P1 ) , (P3 - P1) )

    Author: @fourndo


    .. note::

        Remove all attributes from the GoCAD surface before exporting it!

    """

    fid = open(plFile)
    line = fid.readline()

    # Skip all the lines until the vertices
    VRTX, SEGS = [], []
    while "END" not in line:

        while "VRTX" not in line:
            line = fid.readline()

        vrtx = []
        # Run down all the vertices and save in array
        while np.any(["VRTX" in line, "PVRTX" in line]):
            l_input = re.split(r"[\s*]", line)
            temp = np.array(l_input[2:5])
            vrtx.append(temp.astype(np.float))

            # Read next line
            line = fid.readline()

        VRTX += [np.asarray(vrtx)]

        # Skip lines to the triangles
        while "SEG" not in line:
            line = fid.readline()

        # Run down the list of triangles
        segs = []

        # Run down all the vertices and save in array
        while "SEG" in line:
            l_input = re.split(r"[\s*]", line)
            temp = np.array(l_input[1:3])
            segs.append(temp.astype(np.int))

            # Read next line
            line = fid.readline()

        SEGS += [np.asarray(segs)]

    return VRTX, SEGS


def surface2inds(vrtx, trgl, mesh, boundaries=True, internal=True):
    """"
    Function to read gocad polystructure file and output indexes of
    mesh with in the structure.

    """
    import vtk
    import vtk.util.numpy_support as npsup

    # Adjust the index
    trgl = trgl - 1

    # Make vtk pts
    ptsvtk = vtk.vtkPoints()
    ptsvtk.SetData(npsup.numpy_to_vtk(vrtx, deep=1))

    # Make the polygon connection
    polys = vtk.vtkCellArray()
    for face in trgl:
        poly = vtk.vtkPolygon()
        poly.GetPointIds().SetNumberOfIds(len(face))
        for nrv, vert in enumerate(face):
            poly.GetPointIds().SetId(nrv, vert)
        polys.InsertNextCell(poly)

    # Make the polydata, structure of connections and vrtx
    polyData = vtk.vtkPolyData()
    polyData.SetPoints(ptsvtk)
    polyData.SetPolys(polys)

    # Make implicit func
    ImpDistFunc = vtk.vtkImplicitPolyDataDistance()
    ImpDistFunc.SetInput(polyData)

    # Convert the mesh
    vtkMesh = vtk.vtkRectilinearGrid()
    vtkMesh.SetDimensions(mesh.nNx, mesh.nNy, mesh.nNz)
    vtkMesh.SetXCoordinates(npsup.numpy_to_vtk(mesh.vectorNx, deep=1))
    vtkMesh.SetYCoordinates(npsup.numpy_to_vtk(mesh.vectorNy, deep=1))
    vtkMesh.SetZCoordinates(npsup.numpy_to_vtk(mesh.vectorNz, deep=1))
    # Add indexes
    vtkInd = npsup.numpy_to_vtk(np.arange(mesh.nC), deep=1)
    vtkInd.SetName("Index")
    vtkMesh.GetCellData().AddArray(vtkInd)

    extractImpDistRectGridFilt = vtk.vtkExtractGeometry()  # Object constructor
    extractImpDistRectGridFilt.SetImplicitFunction(ImpDistFunc)  #
    extractImpDistRectGridFilt.SetInputData(vtkMesh)

    if boundaries is True:
        extractImpDistRectGridFilt.ExtractBoundaryCellsOn()

    else:
        extractImpDistRectGridFilt.ExtractBoundaryCellsOff()

    if internal is True:
        extractImpDistRectGridFilt.ExtractInsideOn()

    else:
        extractImpDistRectGridFilt.ExtractInsideOff()

    print("Extracting indices from grid...")
    # Executing the pipe
    extractImpDistRectGridFilt.Update()

    # Get index inside
    insideGrid = extractImpDistRectGridFilt.GetOutput()
    insideGrid = npsup.vtk_to_numpy(insideGrid.GetCellData().GetArray("Index"))

    # Return the indexes inside
    return insideGrid


def download(url, folder=".", overwrite=False, verbose=True):
    """
    Function to download all files stored in a cloud directory

    :param str url: url or list of urls for the file(s) to be downloaded ("https://...")
    :param str folder: folder to where the directory is created and files downloaded (default is the current directory)
    :param bool overwrite: overwrite if a file with the specified name already exists
    :param bool verbose: print out progress
    """

    # Download from cloud
    import urllib
    import os
    import sys

    def rename_path(downloadpath):
        splitfullpath = downloadpath.split(os.path.sep)

        # grab just the filename
        fname = splitfullpath[-1]
        fnamesplit = fname.split(".")
        newname = fnamesplit[0]

        # check if we have already re-numbered
        newnamesplit = newname.split("(")

        # add (num) to the end of the filename
        if len(newnamesplit) == 1:
            num = 1
        else:
            num = int(newnamesplit[-1][:-1])
            num += 1

        newname = "{}({}).{}".format(newnamesplit[0], num, fnamesplit[-1])
        return os.path.sep.join(splitfullpath[:-1] + newnamesplit[:-1] + [newname])

    # grab the correct url retriever
    urlretrieve = urllib.request.urlretrieve

    # ensure we are working with absolute paths and home directories dealt with
    folder = os.path.abspath(os.path.expanduser(folder))

    # make the directory if it doesn't currently exist
    if not os.path.exists(folder):
        os.makedirs(folder)

    if isinstance(url, str):
        filenames = [url.split("/")[-1]]
    elif isinstance(url, list):
        filenames = [u.split("/")[-1] for u in url]

    downloadpath = [os.path.sep.join([folder, f]) for f in filenames]

    # check if the directory already exists
    for i, download in enumerate(downloadpath):
        if os.path.exists(download):
            if overwrite is True:
                if verbose is True:
                    print(f"overwriting {download}")
            elif overwrite is False:
                while os.path.exists is True:
                    download = rename_path(download)

                if verbose is True:
                    print(f"file already exists, new file is called {download}")
                downloadpath[i] = download

    # download files
    urllist = url if isinstance(url, list) else [url]
    for u, f in zip(urllist, downloadpath):
        print(f"Downloading {u}")
        urlretrieve(u, f)
        print("   saved to: " + f)

    print("Download completed!")
    return downloadpath if isinstance(url, list) else downloadpath[0]


def readUBCmagneticsObservations(obs_file):
    """
        Read and write UBC mag file format

        INPUT:
        :param fileName, path to the UBC obs mag file

        OUTPUT:
        :param survey
        :param M, magnetization orentiaton (MI, MD)
    """
    from SimPEG.PF import BaseMag

    fid = open(obs_file)

    # First line has the inclination,declination and amplitude of B0
    line = fid.readline()
    B = np.array(line.split(), dtype=float)

    # Second line has the magnetization orientation and a flag
    line = fid.readline()
    M = np.array(line.split(), dtype=float)

    # Third line has the number of rows
    line = fid.readline()
    ndat = int(line.strip())

    # Pre-allocate space for obsx, obsy, obsz, data, uncert
    line = fid.readline()
    temp = np.array(line.split(), dtype=float)

    d = np.zeros(ndat, dtype=float)
    wd = np.zeros(ndat, dtype=float)
    locXYZ = np.zeros((ndat, 3), dtype=float)

    ii = 0
    while ii < ndat:

        temp = np.array(line.split(), dtype=float)
        if len(temp) > 0:
            locXYZ[ii, :] = temp[:3]

            if len(temp) > 3:
                d[ii] = temp[3]

                if len(temp) == 5:
                    wd[ii] = temp[4]
            ii += 1
        line = fid.readline()

    rxLoc = BaseMag.RxObs(locXYZ)
    srcField = BaseMag.SrcField([rxLoc], param=(B[2], B[0], B[1]))
    survey = BaseMag.LinearSurvey(srcField)
    survey.dobs = d
    survey.std = wd
    return survey, M


def writeUBCmagneticsObservations(filename, survey, d):
    """
    writeUBCobs(filename,B,M,rxLoc,d,wd)

    Function writing an observation file in UBC-MAG3D format.

    INPUT
    filename    : Name of out file including directory
    survey
    flag          : dobs | dpred

    OUTPUT
    Obsfile

    Created on Dec, 27th 2015

    @author: dominiquef
    """

    B = survey.srcField.param

    rxLoc = survey.srcField.rxList[0].locs

    wd = survey.std

    if d.ndim == 2:
        d = d[0]

    data = np.c_[rxLoc, d, wd]
    head = (
        "{:6.2f} {:6.2f} {:6.2f}\n".format(B[1], B[2], B[0])
        + "{:6.2f} {:6.2f} {:6.2f}\n".format(B[1], B[2], 1)
        + "%i" % len(d)
    )
    np.savetxt(
        filename, data, fmt="%e", delimiter=" ", newline="\n", header=head, comments=""
    )


def readUBCgravityObservations(obs_file):
    """
    Read UBC grav file format

    INPUT:
    :param fileName, path to the UBC obs grav file

    OUTPUT:
    :param survey

    """
    from SimPEG.PF import BaseGrav

    fid = open(obs_file)

    # First line has the number of rows
    line = fid.readline()
    ndat = int(line.split()[0])

    # Pre-allocate space for obsx, obsy, obsz, data, uncert
    line = fid.readline()

    d = np.zeros(ndat, dtype=float)
    wd = np.zeros(ndat, dtype=float)
    locXYZ = np.zeros((ndat, 3), dtype=float)

    ii = 0
    while ii < ndat:
        temp = np.array(line.split(), dtype=float)
        if len(temp) > 0:
            locXYZ[ii, :] = temp[:3]
            d[ii] = temp[3]
            wd[ii] = temp[4]
            ii += 1
        line = fid.readline()

    rxLoc = BaseGrav.RxObs(locXYZ)
    srcField = BaseGrav.SrcField([rxLoc])
    survey = BaseGrav.LinearSurvey(srcField)
    survey.dobs = -d
    survey.std = wd
    return survey


def writeUBCgravityObservations(filename, survey, d):
    """
        Write UBC grav file format

        INPUT:
        :param: fileName, path to the UBC obs grav file
        :param: survey Gravity object
        :param: data array

    """
    rxLoc = survey.srcField.rxList[0].locs

    wd = survey.std

    data = np.c_[rxLoc, -d, wd]
    head = "%i" % len(d)
    np.savetxt(
        filename, data, fmt="%e", delimiter=" ", newline="\n", header=head, comments=""
    )


def writeVectorUBC(mesh, fileName, model):
    """
        Writes a vector model associated with a SimPEG TensorMesh
        to a UBC-GIF format model file.

        :param string fileName: File to write to
        :param numpy.ndarray model: The model
    """

    if model.ndim == 1:
        # Catch the standard case that the model is (3*nC,1) instead of (nC,3)
        model = model.reshape((-1, 3), order="F")

    modelMatTR = np.zeros_like(model)

    if isinstance(mesh, TreeMesh):
        ubc_order = mesh._ubc_order

        for ii in range(3):
            modelMatTR[:, ii] = model[ubc_order, ii]

    else:
        for ii in range(3):
            # Reshape model to a matrix
            modelMat = mesh.r(model[:, ii], "CC", "CC", "M")
            # Transpose the axes
            modelMatT = modelMat.transpose((2, 0, 1))
            # Flip z to positive down
            modelMatTR[:, ii] = Utils.mkvc(modelMatT[::-1, :, :])

    # Flip Z for UBC file format
    modelMatTR[:, 2] *= -1

    np.savetxt(fileName, modelMatTR)


def readVectorUBC(mesh, fileName):
    """
        Read a vector model associated with a SimPEG TensorMesh
        to a UBC-GIF format model file.

        :param string fileName: File to write to
        :param numpy.ndarray model: The model
    """

    # f = open(fileName, 'r')
    model = np.loadtxt(fileName)
    vModel = np.zeros((mesh.nC, 3))
    # f.close()
    if isinstance(mesh, TreeMesh):
        ubc_order = mesh._ubc_order

        for ii in range(3):
            vModel[ubc_order, ii] = model[:, ii]
    else:

        for ii in range(3):
            comp = np.reshape(model[:, ii], (mesh.nCz, mesh.nCx, mesh.nCy), order="F")
            comp = comp[::-1, :, :]
            comp = np.transpose(comp, (1, 2, 0))
            vModel[:, ii] = Utils.mkvc(comp)

    # Flip the z vector
    vModel[:, 2] *= -1
    return vModel


class readFTMGdataFile:
    """
    Patch object for holding FTMG data

    """

    #

    def __init__(self, file_path=None):
        if file_path is not None:
            # initiate variables
            self.lines = []
            self.project_name = ""
            self.date = ""
            self.id = ""
            # read the data file
            input_file = open(file_path)
            # extract lines
            cnt = 0
            cnt_line = 0
            header_ids = []
            lines = []
            for dline in input_file:
                # check for header line
                if "/" in dline:
                    if cnt == 1:
                        splt = dline.split("/")
                        self.project_name = splt[1]
                    if cnt == 2:
                        splt = dline.split("/")
                        self.date = splt[1]
                    if cnt == 3:
                        splt = dline.split("/")
                        self.id = splt[1]
                    if cnt == 17:
                        splt = dline.split("/")
                        header_ids = splt[1].split()
                    cnt += 1
                else:
                    if "Line" in dline:
                        if cnt_line == 0:
                            splt = dline.split()
                            new_line = ftmgLine(splt[1])
                            lines.append(new_line)
                            cnt_line += 1
                        else:
                            # print('number of stations: {0} in line: {1}'.format(len(lines[0].stations), lines[0].line_id))
                            self.lines.append(lines[0])
                            splt = dline.split()
                            lines[0] = ftmgLine(splt[1])
                    else:
                        text_line = dline.split()
                        data_dict = {}
                        # check that we have the same amount of header ids
                        # as we do data columns
                        # print("hdr length: {0} & data length {1}".format(len(header_ids), len(text_line)))
                        if len(header_ids) == len(text_line):
                            for idx, ids in enumerate(header_ids):
                                data_dict[ids] = float(text_line[idx])
                                # print(ids, data_dict[ids])
                            # create a stations class
                            lines[0].addStation(ftmgStation(data_obj=data_dict))
                        else:
                            print(
                                "[ERROR] column id and data column mismatch or blank line found!!!"
                            )
            # print('number of stations: {0} in line: {1}'.format(len(lines[0].stations), lines[0].line_id))
            self.lines.append(lines[0])
        else:
            print("Please define an input file!!!!! If you ain't first your last!!!")

    def getUtmLocations(self, limits=None, ground=False):
        """
        Return a numpy array of the easting and northings

        """

        easting = []
        northing = []
        elevation = []
        if limits is None:
            for line in range(len(self.lines)):
                for stn in range(len(self.lines[line].stations)):
                    easting.append(self.lines[line].stations[stn].easting)
                    northing.append(self.lines[line].stations[stn].northing)
                    # see if ground elevation is requested
                    if ground:
                        z_ground = (
                            self.lines[line].stations[stn].alt_gps
                            - self.lines[line].stations[stn].radar
                        )
                        elevation.append(z_ground)
                    else:
                        elevation.append(self.lines[line].stations[stn].alt_gps)
        else:
            for line in range(len(self.lines)):
                for stn in range(len(self.lines[line].stations)):
                    x_stn = self.lines[line].stations[stn].easting
                    y_stn = self.lines[line].stations[stn].northing
                    # check that station is within the limits
                    if limits[0] < x_stn < limits[1] and limits[2] < y_stn < limits[3]:
                        easting.append(self.lines[line].stations[stn].easting)
                        northing.append(self.lines[line].stations[stn].northing)
                        # see if ground elevation is requested
                        if ground:
                            z_ground = (
                                self.lines[line].stations[stn].alt_gps
                                - self.lines[line].stations[stn].radar
                            )
                            elevation.append(z_ground)
                        else:
                            elevation.append(self.lines[line].stations[stn].alt_gps)

        return np.asarray(easting), np.asarray(northing), np.asarray(elevation)

    def getComponentData(self, component=None, limits=None, inducing_field=None):
        output_comp = []

        if component.lower() == "tmi":
            assert (
                inducing_field is not None
            ), "Inducing field parameters must be provided for 'TMI' data"

        # check for data extents requested

        # check thats components are supplied
        if component is not None:
            for line in range(len(self.lines)):
                for stn in range(len(self.lines[line].stations)):

                    if limits is not None:
                        x_stn = self.lines[line].stations[stn].easting
                        y_stn = self.lines[line].stations[stn].northing

                        if not (limits[0] < x_stn < limits[1]) and (
                            limits[2] < y_stn < limits[3]
                        ):
                            continue

                    # check if component was aquired
                    if component in self.lines[line].stations[stn].comps:

                        if component.lower() == "tmi":
                            output_comp.append(
                                self.lines[line].stations[stn].comps[component]
                                - inducing_field[0]
                            )
                        else:
                            output_comp.append(
                                self.lines[line].stations[stn].comps[component]
                            )
        # else:
        #     # check thats components are supplied
        #     if component is not None:
        #         for line in range(len(self.lines)):
        #             for stn in range(len(self.lines[line].stations)):
        #                 x_stn = self.lines[line].stations[stn].easting
        #                 y_stn = self.lines[line].stations[stn].northing
        #                 # check that station is within the limits
        #                 if limits[0] < x_stn < limits[1] and limits[2] < y_stn < limits[3]:
        #                     # check if component was aquired
        #                     if component in self.lines[line].stations[stn].comps:
        #                         output_comp.append(self.lines[line].stations[stn].comps[component])

        return output_comp

    def createFtmgSurvey(self, inducing_field=None, limits=None, force_comp=None):
        """
        Returns a SimPEF survey Object for FTMG to be fed into inversion scheme

        """
        print("TODO")
        # check that we have an inducing field
        from SimPEG.PF import BaseMag

        if inducing_field is not None:
            # check data extents are requested
            if limits is None:
                # check that we have the proper amount of info in array
                if inducing_field.shape[0] == 3:
                    # get observation data
                    x, y, z = self.getUtmLocations()
                    comps = self.getListOfAvailableComponents(
                        format_type="simpeg", force_comp=force_comp
                    )
                    # Create a MAGsurvey
                    rxLoc = np.c_[Utils.mkvc(x.T), Utils.mkvc(y.T), Utils.mkvc(z.T)]
                    Rx = BaseMag.RxObs(rxLoc)
                    srcField = BaseMag.SrcField([Rx], param=inducing_field)
                    survey = BaseMag.LinearSurvey(srcField, components=comps)
                    # create dobs
                    dobs_list = []
                    components = self.getListOfAvailableComponents(
                        force_comp=force_comp
                    )
                    for comp in components:
                        # if comp is not 'TMI':
                        print(f"component added: {comp}")
                        dobs_list.append(
                            self.getComponentData(
                                component=comp, inducing_field=inducing_field
                            )
                        )
                    # now sort the data to SimPEG format
                    dobs = np.vstack(dobs_list)
                    size = dobs.shape[0] * dobs.shape[1]
                    survey.dobs = np.reshape(dobs, size, order="F")

                    return survey

                else:
                    print(
                        "[ERROR] array with Inducing field parameters must be size 3!!!"
                    )
            else:
                # get locations with limits
                x, y, z = self.getUtmLocations(limits=limits)
                comps = self.getListOfAvailableComponents(
                    format_type="simpeg", force_comp=force_comp
                )
                # Create a MAGsurvey
                rxLoc = np.c_[Utils.mkvc(x.T), Utils.mkvc(y.T), Utils.mkvc(z.T)]
                Rx = BaseMag.RxObs(rxLoc)
                srcField = BaseMag.SrcField([Rx], param=inducing_field)
                survey = BaseMag.LinearSurvey(srcField, components=comps)
                # create dobs
                dobs_list = []
                components = self.getListOfAvailableComponents(force_comp=force_comp)
                for comp in components:
                    # if comp is not 'TMI':
                    print(f"component added: {comp}")
                    dobs_list.append(
                        self.getComponentData(component=comp, limits=limits)
                    )
                # now sort the data to SimPEG format
                dobs = np.vstack(dobs_list)
                size = dobs.shape[0] * dobs.shape[1]
                survey.dobs = np.reshape(dobs, size, order="F")

                return survey

        else:
            print(
                "[ERROR] Please include numpy array with Inducing field parameters!!!"
            )

    def getListOfAvailableComponents(
        self, keep_tmi=False, format_type="std", force_comp=None
    ):
        """
        Returns the components collected for given readFTMGdata

        """
        available_components = self.lines[0].stations[0].available_components
        if force_comp is not None:
            available_components = force_comp
        # list_ = []
        # if not keep_tmi:
        #     for comp in available_components:
        #             # if comp is not 'TMI':
        #             list_.append(comp)
        # else:
        list_ = available_components

        if format_type == "simpeg":
            list_bu = list_
            list_ = []
            if "Bxx" in list_bu:
                list_.append("dbx_dx")
            if "Byy" in list_bu:
                list_.append("dby_dy")
            if "Bzz" in list_bu:
                list_.append("dbz_dz")
            if "Bxz" in list_bu:
                list_.append("dbx_dz")
            if "Bxy" in list_bu:
                list_.append("dbx_dy")
            if "Byz" in list_bu:
                list_.append("dby_dz")
            if "Bz" in list_bu:
                list_.append("bz")
            if "Bx" in list_bu:
                list_.append("bx")
            if "By" in list_bu:
                list_.append("by")
            if "TMI" in list_bu:
                list_.append("tmi")
        return list_


class ftmgLine:
    """
    Line class for storing the observations

    """

    def __init__(self, line_id):
        self.stations = []
        self.line_id = line_id

    def addStation(self, station):
        self.stations.append(station)


class ftmgStation:
    """
    station class for hold station observation info

    """

    def __init__(self, data_obj=None):
        self.available_components = []
        self.comps = {}
        self.easting = None
        self.northing = None
        self.lat = None
        self.lon = None
        self.alt_gps = None
        self.radar = None
        self.date = None
        self.gps_sec = None
        self.I1 = None
        self.I2 = None
        if data_obj is not None:
            # assign all availble data
            self.easting = data_obj["x"]
            self.northing = data_obj["y"]
            self.lat = data_obj["lat"]
            self.lon = data_obj["lon"]
            self.alt_gps = data_obj["alt_gps"]
            self.radar = data_obj["radar"]
            self.date = data_obj["date"]
            self.gps_sec = data_obj["GPS_sec"]
            if "Bxx" in data_obj:
                # self.Bxx = data_obj['Bxx']
                self.comps["Bxx"] = data_obj["Bxx"]
                self.available_components.append("Bxx")
            if "Bxy" in data_obj:
                self.comps["Bxy"] = data_obj["Bxy"]
                # self.Bxy = data_obj['Bxy']
                self.available_components.append("Bxy")
            if "Bxz" in data_obj:
                self.comps["Bxz"] = data_obj["Bxz"]
                # self.Bxz = data_obj['Bxz']
                self.available_components.append("Bxz")
            if "Byy" in data_obj:
                self.comps["Byy"] = data_obj["Byy"]
                # self.Byy = data_obj['Byy']
                self.available_components.append("Byy")
            if "Byz" in data_obj:
                self.comps["Byz"] = data_obj["Byz"]
                # self.Byz = data_obj['Byz']
                self.available_components.append("Byz")
            if "Bzz" in data_obj:
                self.comps["Bzz"] = data_obj["Bzz"]
                # self.Bzz = data_obj['Bzz']
                self.available_components.append("Bzz")
            if "Bx" in data_obj:
                self.comps["Bx"] = data_obj["Bx"]
                # self.Bx = data_obj['Bx']
                self.available_components.append("Bx")
            if "By" in data_obj:
                self.comps["By"] = data_obj["By"]
                # self.By = data_obj['By']
                self.available_components.append("By")
            if "Bz" in data_obj:
                self.comps["Bz"] = data_obj["Bz"]
                # self.Bz = data_obj['Bz']
                self.available_components.append("Bz")
            if "TMI" in data_obj:
                self.comps["TMI"] = data_obj["TMI"]
                # self.tmi = data_obj['TMI']
                self.available_components.append("TMI")
            self.I1 = data_obj["I1"]
            self.I2 = data_obj["I2"]



import properties
import numpy as np
from scipy.constants import mu_0
import warnings

from SimPEG.Utils import Zero
from .. import Survey, Problem, Utils

from .. import Utils as emutils
from ..Base import BaseEMSrc


class BaseFDEMSrc(BaseEMSrc):
    """
    Base source class for FDEM Survey
    """

    freq = properties.Float("frequency of the source", min=0, required=True)

    _ePrimary = None
    _bPrimary = None
    _hPrimary = None
    _jPrimary = None

    def __init__(self, rxList, **kwargs):
        super().__init__(rxList, **kwargs)

    def bPrimary(self, prob):
        """
        Primary magnetic flux density

        :param BaseFDEMProblem prob: FDEM Problem
        :rtype: numpy.ndarray
        :return: primary magnetic flux density
        """
        if self._bPrimary is None:
            return Zero()
        return self._bPrimary

    def bPrimaryDeriv(self, prob, v, adjoint=False):
        """
        Derivative of the primary magnetic flux density

        :param BaseFDEMProblem prob: FDEM Problem
        :param numpy.ndarray v: vector
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: primary magnetic flux density
        """
        return Zero()

    def hPrimary(self, prob):
        """
        Primary magnetic field

        :param BaseFDEMProblem prob: FDEM Problem
        :rtype: numpy.ndarray
        :return: primary magnetic field
        """
        if self._hPrimary is None:
            return Zero()
        return self._hPrimary

    def hPrimaryDeriv(self, prob, v, adjoint=False):
        """
        Derivative of the primary magnetic field

        :param BaseFDEMProblem prob: FDEM Problem
        :param numpy.ndarray v: vector
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: primary magnetic flux density
        """
        return Zero()

    def ePrimary(self, prob):
        """
        Primary electric field

        :param BaseFDEMProblem prob: FDEM Problem
        :rtype: numpy.ndarray
        :return: primary electric field
        """
        if self._ePrimary is None:
            return Zero()
        return self._ePrimary

    def ePrimaryDeriv(self, prob, v, adjoint=False):
        """
        Derivative of the primary electric field

        :param BaseFDEMProblem prob: FDEM Problem
        :param numpy.ndarray v: vector
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: primary magnetic flux density
        """
        return Zero()

    def jPrimary(self, prob):
        """
        Primary current density

        :param BaseFDEMProblem prob: FDEM Problem
        :rtype: numpy.ndarray
        :return: primary current density
        """
        if self._jPrimary is None:
            return Zero()
        return self._jPrimary

    def jPrimaryDeriv(self, prob, v, adjoint=False):
        """
        Derivative of the primary current density

        :param BaseFDEMProblem prob: FDEM Problem
        :param numpy.ndarray v: vector
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: primary magnetic flux density
        """
        return Zero()


class RawVec_e(BaseFDEMSrc):
    """
    RawVec electric source. It is defined by the user provided vector s_e

    :param list rxList: receiver list
    :param float freq: frequency
    :param numpy.array s_e: electric source term
    :param bool integrate: Integrate the source term (multiply by Me) [False]
    """

    def __init__(self, rxList, freq, s_e, **kwargs):
        self._s_e = np.array(s_e, dtype=complex)
        self.freq = freq

        super(BaseFDEMSrc, self).__init__(rxList, **kwargs)

    def s_e(self, prob):
        """
        Electric source term

        :param BaseFDEMProblem prob: FDEM Problem
        :rtype: numpy.ndarray
        :return: electric source term on mesh
        """
        if prob._formulation == "EB" and self.integrate is True:
            return prob.Me * self._s_e
        return self._s_e


class RawVec_m(BaseFDEMSrc):
    """
    RawVec magnetic source. It is defined by the user provided vector s_m

    :param float freq: frequency
    :param rxList: receiver list
    :param numpy.array s_m: magnetic source term
    :param bool integrate: Integrate the source term (multiply by Me) [False]
    """

    def __init__(self, rxList, freq, s_m, **kwargs):
        self._s_m = np.array(s_m, dtype=complex)
        self.freq = freq
        super().__init__(rxList, **kwargs)

    def s_m(self, prob):
        """
        Magnetic source term

        :param BaseFDEMProblem prob: FDEM Problem
        :rtype: numpy.ndarray
        :return: magnetic source term on mesh
        """
        if prob._formulation == "HJ" and self.integrate is True:
            return prob.Me * self._s_m
        return self._s_m


class RawVec(BaseFDEMSrc):
    """
    RawVec source. It is defined by the user provided vectors s_m, s_e

    :param rxList: receiver list
    :param float freq: frequency
    :param numpy.array s_m: magnetic source term
    :param numpy.array s_e: electric source term
    :param bool integrate: Integrate the source term (multiply by Me) [False]
    """

    def __init__(self, rxList, freq, s_m, s_e, **kwargs):
        self._s_m = np.array(s_m, dtype=complex)
        self._s_e = np.array(s_e, dtype=complex)
        self.freq = freq
        super().__init__(rxList, **kwargs)

    def s_m(self, prob):
        """
        Magnetic source term

        :param BaseFDEMProblem prob: FDEM Problem
        :rtype: numpy.ndarray
        :return: magnetic source term on mesh
        """
        if prob._formulation == "HJ" and self.integrate is True:
            return prob.Me * self._s_m
        return self._s_m

    def s_e(self, prob):
        """
        Electric source term

        :param BaseFDEMProblem prob: FDEM Problem
        :rtype: numpy.ndarray
        :return: electric source term on mesh
        """
        if prob._formulation == "EB" and self.integrate is True:
            return prob.Me * self._s_e
        return self._s_e


class MagDipole(BaseFDEMSrc):
    """
    Point magnetic dipole source calculated by taking the curl of a magnetic
    vector potential. By taking the discrete curl, we ensure that the magnetic
    flux density is divergence free (no magnetic monopoles!).

    This approach uses a primary-secondary in frequency. Here we show the
    derivation for E-B formulation noting that similar steps are followed for
    the H-J formulation.

    .. math::
        \\mathbf{C} \\mathbf{e} + i \\omega \\mathbf{b} = \\mathbf{s_m} \\\\
        {\\mathbf{C}^T \\mathbf{M_{\\mu^{-1}}^f} \\mathbf{b} -
        \\mathbf{M_{\\sigma}^e} \\mathbf{e} = \\mathbf{s_e}}

    We split up the fields and :math:`\\mu^{-1}` into primary
    (:math:`\\mathbf{P}`) and secondary (:math:`\\mathbf{S}`) components

    - :math:`\\mathbf{e} = \\mathbf{e^P} + \\mathbf{e^S}`
    - :math:`\\mathbf{b} = \\mathbf{b^P} + \\mathbf{b^S}`
    - :math:`\\boldsymbol{\\mu}^{\\mathbf{-1}} =
      \\boldsymbol{\\mu}^{\\mathbf{-1}^\\mathbf{P}} +
      \\boldsymbol{\\mu}^{\\mathbf{-1}^\\mathbf{S}}`

    and define a zero-frequency primary problem, noting that the source is
    generated by a divergence free electric current

    .. math::
        \\mathbf{C} \\mathbf{e^P} = \\mathbf{s_m^P} = 0 \\\\
        {\\mathbf{C}^T \\mathbf{{M_{\\mu^{-1}}^f}^P} \\mathbf{b^P} -
        \\mathbf{M_{\\sigma}^e} \\mathbf{e^P} = \\mathbf{M^e} \\mathbf{s_e^P}}

    Since :math:`\\mathbf{e^P}` is curl-free, divergence-free, we assume that
    there is no constant field background, the :math:`\\mathbf{e^P} = 0`, so our
    primary problem is

    .. math::
        \\mathbf{e^P} =  0 \\\\
            {\\mathbf{C}^T \\mathbf{{M_{\\mu^{-1}}^f}^P} \\mathbf{b^P} =
            \\mathbf{s_e^P}}

    Our secondary problem is then

    .. math::
        \\mathbf{C} \\mathbf{e^S} + i \\omega \\mathbf{b^S} =
        - i \\omega \\mathbf{b^P} \\\\
        {\\mathbf{C}^T \\mathbf{M_{\\mu^{-1}}^f} \\mathbf{b^S} -
        \\mathbf{M_{\\sigma}^e} \\mathbf{e^S} =
        -\\mathbf{C}^T \\mathbf{{M_{\\mu^{-1}}^f}^S} \\mathbf{b^P}}

    :param list rxList: receiver list
    :param float freq: frequency
    :param numpy.ndarray loc: source location
        (ie: :code:`np.r_[xloc,yloc,zloc]`)
    :param string orientation: 'X', 'Y', 'Z'
    :param float moment: magnetic dipole moment
    :param float mu: background magnetic permeability

    """

    moment = properties.Float("dipole moment of the transmitter", default=1.0, min=0.0)
    mu = properties.Float("permeability of the background", default=mu_0, min=0.0)
    orientation = properties.Vector3(
        "orientation of the source", default="Z", length=1.0, required=True
    )
    freq = properties.Float("frequency of the source (Hz)", required=True)
    loc = properties.Vector3("location of the source", default=np.r_[0.0, 0.0, 0.0])

    def __init__(self, rxList, freq, loc, **kwargs):

        super().__init__(rxList, **kwargs)

        self.freq = freq
        self.loc = loc

    @properties.validator("orientation")
    def _warn_non_axis_aligned_sources(self, change):
        value = change["value"]
        axaligned = [
            True
            for vec in [
                np.r_[1.0, 0.0, 0.0],
                np.r_[0.0, 1.0, 0.0],
                np.r_[0.0, 0.0, 1.0],
            ]
            if np.all(value == vec)
        ]
        if len(axaligned) != 1:
            warnings.warn(
                "non-axes aligned orientations {} are not rigorously"
                " tested".format(value)
            )

    def _srcFct(self, obsLoc, component):
        return emutils.MagneticDipoleVectorPotential(
            self.loc,
            obsLoc,
            component,
            mu=self.mu,
            moment=self.moment,
            orientation=self.orientation,
        )

    def bPrimary(self, prob):
        """
        The primary magnetic flux density from a magnetic vector potential

        :param BaseFDEMProblem prob: FDEM problem
        :rtype: numpy.ndarray
        :return: primary magnetic field
        """
        formulation = prob._formulation

        if formulation == "EB":
            gridX = prob.mesh.gridEx
            gridY = prob.mesh.gridEy
            gridZ = prob.mesh.gridEz
            C = prob.mesh.edgeCurl

        elif formulation == "HJ":
            gridX = prob.mesh.gridFx
            gridY = prob.mesh.gridFy
            gridZ = prob.mesh.gridFz
            C = prob.mesh.edgeCurl.T

        if prob.mesh._meshType == "CYL":
            if not prob.mesh.isSymmetric:
                # TODO ?
                raise NotImplementedError("Non-symmetric cyl mesh not implemented yet!")
            assert np.linalg.norm(self.orientation - np.r_[0.0, 0.0, 1.0]) < 1e-6, (
                "for cylindrical symmetry, the dipole must be "
                "oriented in the Z direction"
            )
            a = self._srcFct(gridY, "y")
        else:
            ax = self._srcFct(gridX, "x")
            ay = self._srcFct(gridY, "y")
            az = self._srcFct(gridZ, "z")
            a = np.concatenate((ax, ay, az))

        return C * a

    def hPrimary(self, prob):
        """
        The primary magnetic field from a magnetic vector potential

        :param BaseFDEMProblem prob: FDEM problem
        :rtype: numpy.ndarray
        :return: primary magnetic field
        """
        b = self.bPrimary(prob)
        return 1.0 / self.mu * b

    def s_m(self, prob):
        """
        The magnetic source term

        :param BaseFDEMProblem prob: FDEM problem
        :rtype: numpy.ndarray
        :return: primary magnetic field
        """

        b_p = self.bPrimary(prob)
        if prob._formulation == "HJ":
            b_p = prob.Me * b_p
        return -1j * emutils.omega(self.freq) * b_p

    def s_e(self, prob):
        """
        The electric source term

        :param BaseFDEMProblem prob: FDEM problem
        :rtype: numpy.ndarray
        :return: primary magnetic field
        """

        if all(np.r_[self.mu] == np.r_[prob.mu]):
            return Zero()
        else:
            formulation = prob._formulation

            if formulation == "EB":
                mui_s = prob.mui - 1.0 / self.mu
                MMui_s = prob.mesh.getFaceInnerProduct(mui_s)
                C = prob.mesh.edgeCurl
            elif formulation == "HJ":
                mu_s = prob.mu - self.mu
                MMui_s = prob.mesh.getEdgeInnerProduct(mu_s, invMat=True)
                C = prob.mesh.edgeCurl.T

            return -C.T * (MMui_s * self.bPrimary(prob))

    def s_eDeriv(self, prob, v, adjoint=False):
        if not hasattr(prob, "muMap") or not hasattr(prob, "muiMap"):
            return Zero()
        else:
            formulation = prob._formulation

            if formulation == "EB":
                mui_s = prob.mui - 1.0 / self.mu
                MMui_sDeriv = (
                    prob.mesh.getFaceInnerProductDeriv(mui_s)(self.bPrimary(prob))
                    * prob.muiDeriv
                )
                C = prob.mesh.edgeCurl

                if adjoint:
                    return -MMui_sDeriv.T * (C * v)

                return -C.T * (MMui_sDeriv * v)

            elif formulation == "HJ":
                return Zero()
                # raise NotImplementedError
                mu_s = prob.mu - self.mu
                MMui_s = prob.mesh.getEdgeInnerProduct(mu_s, invMat=True)
                C = prob.mesh.edgeCurl.T

                return -C.T * (MMui_s * self.bPrimary(prob))


class MagDipole_Bfield(MagDipole):

    """
    Point magnetic dipole source calculated with the analytic solution for the
    fields from a magnetic dipole. No discrete curl is taken, so the magnetic
    flux density may not be strictly divergence free.

    This approach uses a primary-secondary in frequency in the same fashion as
    the MagDipole.

    :param list rxList: receiver list
    :param float freq: frequency
    :param numpy.ndarray loc: source location (ie:
                              :code:`np.r_[xloc,yloc,zloc]`)
    :param string orientation: 'X', 'Y', 'Z'
    :param float moment: magnetic dipole moment
    :param float mu: background magnetic permeability
    """

    def __init__(self, rxList, freq, loc, **kwargs):
        super().__init__(rxList, freq=freq, loc=loc, **kwargs)

    def _srcFct(self, obsLoc, component):
        return emutils.MagneticDipoleFields(
            self.loc,
            obsLoc,
            component,
            mu=self.mu,
            moment=self.moment,
            orientation=self.orientation,
        )

    def bPrimary(self, prob):
        """
        The primary magnetic flux density from the analytic solution for
        magnetic fields from a dipole

        :param BaseFDEMProblem prob: FDEM problem
        :rtype: numpy.ndarray
        :return: primary magnetic field
        """

        formulation = prob._formulation

        if formulation == "EB":
            gridX = prob.mesh.gridFx
            gridY = prob.mesh.gridFy
            gridZ = prob.mesh.gridFz

        elif formulation == "HJ":
            gridX = prob.mesh.gridEx
            gridY = prob.mesh.gridEy
            gridZ = prob.mesh.gridEz

        srcfct = emutils.MagneticDipoleFields
        if prob.mesh._meshType == "CYL":
            if not prob.mesh.isSymmetric:
                # TODO ?
                raise NotImplementedError("Non-symmetric cyl mesh not implemented yet!")
            bx = srcfct(self.loc, gridX, "x", mu=self.mu, moment=self.moment)
            bz = srcfct(self.loc, gridZ, "z", mu=self.mu, moment=self.moment)
            b = np.concatenate((bx, bz))
        else:
            bx = srcfct(self.loc, gridX, "x", mu=self.mu, moment=self.moment)
            by = srcfct(self.loc, gridY, "y", mu=self.mu, moment=self.moment)
            bz = srcfct(self.loc, gridZ, "z", mu=self.mu, moment=self.moment)
            b = np.concatenate((bx, by, bz))

        return Utils.mkvc(b)


class CircularLoop(MagDipole):
    """
    Circular loop magnetic source calculated by taking the curl of a magnetic
    vector potential. By taking the discrete curl, we ensure that the magnetic
    flux density is divergence free (no magnetic monopoles!).

    This approach uses a primary-secondary in frequency in the same fashion as
    the MagDipole.

    :param list rxList: receiver list
    :param float freq: frequency
    :param numpy.ndarray loc: source location
        (ie: :code:`np.r_[xloc,yloc,zloc]`)
    :param string orientation: 'X', 'Y', 'Z'
    :param float moment: magnetic dipole moment
    :param float mu: background magnetic permeability
    """

    radius = properties.Float("radius of the loop", default=1.0, min=0.0)

    def __init__(self, rxList, freq, loc, **kwargs):
        super().__init__(rxList, freq, loc, **kwargs)

    def _srcFct(self, obsLoc, component):
        return emutils.MagneticLoopVectorPotential(
            self.loc,
            obsLoc,
            component,
            mu=self.mu,
            radius=self.radius,
            orientation=self.orientation,
        )


class PrimSecSigma(BaseFDEMSrc):
    def __init__(self, rxList, freq, sigBack, ePrimary, **kwargs):
        self.sigBack = sigBack

        BaseFDEMSrc.__init__(self, rxList, freq=freq, _ePrimary=ePrimary, **kwargs)

    def s_e(self, prob):
        return (
            prob.MeSigma - prob.mesh.getEdgeInnerProduct(self.sigBack)
        ) * self.ePrimary(prob)

    def s_eDeriv(self, prob, v, adjoint=False):
        if adjoint:
            return prob.MeSigmaDeriv(self.ePrimary(prob)).T * v
        return prob.MeSigmaDeriv(self.ePrimary(prob)) * v


class PrimSecMappedSigma(BaseFDEMSrc):

    """
    Primary-Secondary Source in which a mapping is provided to put the current
    model onto the primary mesh. This is solved on every model update.
    There are a lot of layers to the derivatives here!

    **Required**
    :param list rxList: Receiver List
    :param float freq: frequency
    :param BaseFDEMProblem primaryProblem: FDEM primary problem
    :param SurveyFDEM primarySurvey: FDEM primary survey

    **Optional**
    :param Mapping map2meshSecondary: mapping current model to act as primary
    model on the secondary mesh
    """

    def __init__(
        self,
        rxList,
        freq,
        primaryProblem,
        primarySurvey,
        map2meshSecondary=None,
        **kwargs
    ):

        self.primaryProblem = primaryProblem
        self.primarySurvey = primarySurvey

        if self.primaryProblem.ispaired is False:
            self.primaryProblem.pair(self.primarySurvey)

        self.map2meshSecondary = map2meshSecondary

        BaseFDEMSrc.__init__(self, rxList, freq=freq, **kwargs)

    def _ProjPrimary(self, prob, locType, locTypeTo):
        # TODO: if meshes have not changed, store the projection
        # if getattr(self, '__ProjPrimary', None) is None:

        # TODO: implement for HJ formulation
        if prob._formulation == "EB":
            pass
        else:
            raise NotImplementedError(
                "PrimSecMappedSigma Source has not been implemented for {} "
                "formulation".format(prob._formulation)
            )

        # TODO: only set up for tensot meshes (Tree meshes should be easy/done)
        # but have not been tried or tested.
        assert prob.mesh._meshType in [
            "TENSOR"
        ], "PrimSecMappedSigma source has not been implemented for {}".format(
            prob.mesh._meshType
        )

        # if EB formulation, interpolate E, elif HJ interpolate J
        # if self.primaryProblem._formulation == 'EB':
        #     locType = 'E'
        # elif self.primaryProblem._formulation == 'HJ':
        #     locType = 'F'

        # get interpolation mat from primary mesh to secondary mesh
        if self.primaryProblem.mesh._meshType == "CYL":
            return self.primaryProblem.mesh.getInterpolationMatCartMesh(
                prob.mesh, locType=locType, locTypeTo=locTypeTo
            )
        return self.primaryProblem.mesh.getInterploationMat(
            prob.mesh, locType=locType, locTypeTo=locTypeTo
        )

        # return self.__ProjPrimary

    def _primaryFields(self, prob, fieldType=None, f=None):
        # TODO: cache and check if prob.curModel has changed

        if f is None:
            f = self.primaryProblem.fields(prob.model)

        if fieldType is not None:
            return f[:, fieldType]
        return f

    def _primaryFieldsDeriv(self, prob, v, adjoint=False, f=None):
        # TODO: this should not be hard-coded for j
        # jp = self._primaryFields(prob)[:,'j']

        # TODO: pull apart Jvec so that don't have to copy paste this code in
        # A = self.primaryProblem.getA(self.freq)
        # Ainv = self.primaryProblem.Solver(A, **self.primaryProblem.solverOpts) # create the concept of Ainv (actually a solve)

        if f is None:
            f = self._primaryFields(prob.sigma, f=f)

        freq = self.freq

        A = self.primaryProblem.getA(freq)
        src = self.primarySurvey.srcList[0]
        u_src = Utils.mkvc(f[src, self.primaryProblem._solutionType])

        if adjoint is True:
            Jtv = np.zeros(prob.sigmaMap.nP, dtype=complex)
            ATinv = self.primaryProblem.Solver(A.T, **self.primaryProblem.solverOpts)
            df_duTFun = getattr(
                f,
                "_{}Deriv".format(
                    "e" if self.primaryProblem._formulation == "EB" else "j"
                ),
                None,
            )
            df_duT, df_dmT = df_duTFun(src, None, v, adjoint=True)

            ATinvdf_duT = ATinv * df_duT

            dA_dmT = self.primaryProblem.getADeriv(
                freq, u_src, ATinvdf_duT, adjoint=True
            )
            dRHS_dmT = self.primaryProblem.getRHSDeriv(
                freq, src, ATinvdf_duT, adjoint=True
            )

            du_dmT = -dA_dmT + dRHS_dmT

            Jtv += df_dmT + du_dmT

            ATinv.clean()

            return Utils.mkvc(Jtv)

        # create the concept of Ainv (actually a solve)
        Ainv = self.primaryProblem.Solver(A, **self.primaryProblem.solverOpts)

        # for src in self.survey.getSrcByFreq(freq):
        dA_dm_v = self.primaryProblem.getADeriv(freq, u_src, v)
        dRHS_dm_v = self.primaryProblem.getRHSDeriv(freq, src, v)
        du_dm_v = Ainv * (-dA_dm_v + dRHS_dm_v)

        # if self.primaryProblem._formulation == 'EB':
        df_dmFun = getattr(
            f,
            "_{}Deriv".format("e" if self.primaryProblem._formulation == "EB" else "j"),
            None,
        )
        # elif self.primaryProblem._formulation == 'HJ':
        #     df_dmFun = getattr(f, '_{0}Deriv'.format('j'), None)
        df_dm_v = df_dmFun(src, du_dm_v, v, adjoint=False)
        # Jv[src, rx] = rx.evalDeriv(src, self.mesh, f, df_dm_v)
        Ainv.clean()

        return df_dm_v

        # return self.primaryProblem.Jvec(prob.curModel, v, f=f)

    def ePrimary(self, prob, f=None):
        if f is None:
            f = self._primaryFields(prob)

        if self.primaryProblem._formulation == "EB":
            ep = self._ProjPrimary(prob, "E", "E") * f[:, "e"]
        elif self.primaryProblem._formulation == "HJ":
            ep = self._ProjPrimary(prob, "F", "E") * (
                self.primaryProblem.MfI * (self.primaryProblem.MfRho * f[:, "j"])
            )

        return Utils.mkvc(ep)

    def ePrimaryDeriv(self, prob, v, adjoint=False, f=None):

        if f is None:
            f = self._primaryFields(prob)

        # if adjoint is True:
        #     raise NotImplementedError
        if self.primaryProblem._formulation == "EB":
            if adjoint is True:
                epDeriv = self._primaryFieldsDeriv(
                    prob,
                    (self._ProjPrimary(prob, "E", "E").T * v),
                    f=f,
                    adjoint=adjoint,
                )
            else:
                epDeriv = self._ProjPrimary(prob, "E", "E") * self._primaryFieldsDeriv(
                    prob, v, f=f
                )
        elif self.primaryProblem._formulation == "HJ":
            if adjoint is True:
                PTv = self.primaryProblem.MfI.T * (
                    self._ProjPrimary(prob, "F", "E").T * v
                )
                epDeriv = self.primaryProblem.MfRhoDeriv(
                    f[:, "j"]
                ).T * PTv + self._primaryFieldsDeriv(
                    prob, self.primaryProblem.MfRho.T * PTv, adjoint=adjoint, f=f
                )
                # epDeriv =(

                #     (self.primaryProblem.MfI.T * PTv)
                #     )
            else:
                epDeriv = self._ProjPrimary(prob, "F", "E") * (
                    self.primaryProblem.MfI
                    * (
                        (self.primaryProblem.MfRhoDeriv(f[:, "j"]) * v)
                        + (
                            self.primaryProblem.MfRho
                            * self._primaryFieldsDeriv(prob, v, f=f)
                        )
                    )
                )

        return Utils.mkvc(epDeriv)

    def bPrimary(self, prob, f=None):
        if f is None:
            f = self._primaryFields(prob)

        if self.primaryProblem._formulation == "EB":
            bp = self._ProjPrimary(prob, "F", "F") * f[:, "b"]
        elif self.primaryProblem._formulation == "HJ":
            bp = self._ProjPrimary(prob, "E", "F") * (
                self.primaryProblem.MeI * (self.primaryProblem.MeMu * f[:, "h"])
            )

        return Utils.mkvc(bp)

    def s_e(self, prob, f=None):
        sigmaPrimary = self.map2meshSecondary * prob.model

        return Utils.mkvc(
            (prob.MeSigma - prob.mesh.getEdgeInnerProduct(sigmaPrimary))
            * self.ePrimary(prob, f=f)
        )

    def s_eDeriv(self, prob, v, adjoint=False):

        sigmaPrimary = self.map2meshSecondary * prob.model
        sigmaPrimaryDeriv = self.map2meshSecondary.deriv(prob.model)

        f = self._primaryFields(prob)
        ePrimary = self.ePrimary(prob, f=f)

        if adjoint is True:
            return (
                prob.MeSigmaDeriv(ePrimary).T * v
                - (
                    sigmaPrimaryDeriv.T
                    * prob.mesh.getEdgeInnerProductDeriv(sigmaPrimary)(ePrimary).T
                    * v
                )
                + self.ePrimaryDeriv(
                    prob,
                    (prob.MeSigma - prob.mesh.getEdgeInnerProduct(sigmaPrimary)).T * v,
                    adjoint=adjoint,
                    f=f,
                )
            )

        return (
            prob.MeSigmaDeriv(ePrimary) * v
            - prob.mesh.getEdgeInnerProductDeriv(sigmaPrimary)(ePrimary)
            * (sigmaPrimaryDeriv * v)
            + (prob.MeSigma - prob.mesh.getEdgeInnerProduct(sigmaPrimary))
            * self.ePrimaryDeriv(prob, v, adjoint=adjoint, f=f)
        )



import numpy as np

import SimPEG as simpeg
from SimPEG.Utils import meshTensor
from SimPEG.EM.NSEM.RxNSEM import Point_impedance1D, Point_impedance3D, Point_tipper3D
from SimPEG.EM.NSEM.SurveyNSEM import Survey
from SimPEG.EM.NSEM.SrcNSEM import Planewave_xy_1Dprimary, Planewave_xy_1DhomotD
from SimPEG.EM.NSEM.ProblemNSEM import Problem3D_ePrimSec
from .dataUtils import appResPhs

np.random.seed(1100)
# Define the tolerances
TOLr = 5e-2
TOLp = 5e-2


def getAppResPhs(NSEMdata):
    # Make impedance

    zList = []
    for src in NSEMdata.survey.srcList:
        zc = [src.freq]
        for rx in src.rxList:
            if "imag" in rx.component:
                m = 1j
            else:
                m = 1
            zc.append(m * NSEMdata[src, rx])
        zList.append(zc)
    return [
        appResPhs(zList[i][0], np.sum(zList[i][1:3])) for i in np.arange(len(zList))
    ]


def setup1DSurvey(sigmaHalf, tD=False, structure=False):

    # Frequency
    nFreq = 33
    freqs = np.logspace(3, -3, nFreq)
    # Make the mesh
    ct = 5
    air = meshTensor([(ct, 25, 1.3)])
    # coreT0 = meshTensor([(ct,15,1.2)])
    # coreT1 = np.kron(meshTensor([(coreT0[-1],15,1.3)]),np.ones((7,)))
    core = np.concatenate(
        (np.kron(meshTensor([(ct, 15, -1.2)]), np.ones((10,))), meshTensor([(ct, 20)]))
    )
    bot = meshTensor([(core[0], 20, -1.3)])
    x0 = -np.array([np.sum(np.concatenate((core, bot)))])
    m1d = simpeg.Mesh.TensorMesh([np.concatenate((bot, core, air))], x0=x0)
    # Make the model
    sigma = np.zeros(m1d.nC) + sigmaHalf
    sigma[m1d.gridCC > 0] = 1e-8
    sigmaBack = sigma.copy()
    # Add structure
    if structure:
        shallow = (m1d.gridCC < -200) * (m1d.gridCC > -600)
        deep = (m1d.gridCC < -3000) * (m1d.gridCC > -5000)
        sigma[shallow] = 1
        sigma[deep] = 0.1

    rxList = []
    for rxType in ["z1d", "z1d"]:
        rxList.append(Point_impedance1D(simpeg.mkvc(np.array([0.0]), 2).T, "real"))
        rxList.append(Point_impedance1D(simpeg.mkvc(np.array([0.0]), 2).T, "imag"))
    # Source list
    srcList = []
    if tD:
        for freq in freqs:
            srcList.append(Planewave_xy_1DhomotD(rxList, freq))
    else:
        for freq in freqs:
            srcList.append(Planewave_xy_1Dprimary(rxList, freq))

    survey = Survey(srcList)
    return (survey, sigma, sigmaBack, m1d)


def setupSimpegNSEM_ePrimSec(inputSetup, comp="Imp", singleFreq=False, expMap=True):

    M, freqs, sig, sigBG, rx_loc = inputSetup
    # Make a receiver list
    rxList = []
    if comp == "All":
        rx_type_list = ["xx", "xy", "yx", "yy", "zx", "zy"]
    elif comp == "Imp":
        rx_type_list = ["xx", "xy", "yx", "yy"]
    elif comp == "Tip":
        rx_type_list = ["zx", "zy"]
    else:
        rx_type_list = [comp]

    for rx_type in rx_type_list:
        if rx_type in ["xx", "xy", "yx", "yy"]:
            rxList.append(Point_impedance3D(rx_loc, rx_type, "real"))
            rxList.append(Point_impedance3D(rx_loc, rx_type, "imag"))
        if rx_type in ["zx", "zy"]:
            rxList.append(Point_tipper3D(rx_loc, rx_type, "real"))
            rxList.append(Point_tipper3D(rx_loc, rx_type, "imag"))

    # Source list
    srcList = []

    if singleFreq:
        srcList.append(Planewave_xy_1Dprimary(rxList, singleFreq))
    else:
        for freq in freqs:
            srcList.append(Planewave_xy_1Dprimary(rxList, freq))
    # Survey NSEM
    survey = Survey(srcList)

    # Setup the problem object
    sigma1d = M.r(sigBG, "CC", "CC", "M")[0, 0, :]

    if expMap:
        problem = Problem3D_ePrimSec(M, sigmaPrimary=np.log(sigma1d))
        problem.sigmaMap = simpeg.Maps.ExpMap(problem.mesh)
        problem.model = np.log(sig)
    else:
        problem = Problem3D_ePrimSec(M, sigmaPrimary=sigma1d)
        problem.sigmaMap = simpeg.Maps.IdentityMap(problem.mesh)
        problem.model = sig
    problem.pair(survey)
    problem.verbose = False
    try:
        from pymatsolver import Pardiso

        problem.Solver = Pardiso
    except:
        pass

    return (survey, problem)


def getInputs():
    """
    Function that returns Mesh, freqs, rx_loc, elev.
    """
    # Make a mesh
    M = simpeg.Mesh.TensorMesh(
        [
            [(200, 6, -1.5), (200.0, 4), (200, 6, 1.5)],
            [(200, 6, -1.5), (200.0, 4), (200, 6, 1.5)],
            [(200, 8, -1.5), (200.0, 8), (200, 8, 1.5)],
        ],
        x0=["C", "C", "C"],
    )  # Setup the model
    # Set the frequencies
    freqs = np.logspace(1, -3, 5)
    elev = 0

    # Setup the the survey object
    # Receiver locations
    rx_x, rx_y = np.meshgrid(np.arange(-350, 350, 200), np.arange(-350, 350, 200))
    rx_loc = np.hstack(
        (
            simpeg.Utils.mkvc(rx_x, 2),
            simpeg.Utils.mkvc(rx_y, 2),
            elev + np.zeros((np.prod(rx_x.shape), 1)),
        )
    )

    return M, freqs, rx_loc, elev


def random(conds):
    """ Returns a random model based on the inputs"""
    M, freqs, rx_loc, elev = getInputs()

    # Backround
    sigBG = np.ones(M.nC) * conds
    # Add randomness to the model (10% of the value).
    sig = np.exp(np.log(sigBG) + np.random.randn(M.nC) * (conds) * 1e-1)

    return (M, freqs, sig, sigBG, rx_loc)


def halfSpace(conds):
    """ Returns a halfspace model based on the inputs"""
    M, freqs, rx_loc, elev = getInputs()

    # Model
    ccM = M.gridCC
    # conds = [1e-2]
    groundInd = ccM[:, 2] < elev
    sig = np.zeros(M.nC) + 1e-8
    sig[groundInd] = conds
    # Set the background, not the same as the model
    sigBG = np.zeros(M.nC) + 1e-8
    sigBG[groundInd] = conds

    return (M, freqs, sig, sigBG, rx_loc)


def blockInhalfSpace(conds):
    """ Returns a block in a halfspace model based on the inputs"""
    M, freqs, rx_loc, elev = getInputs()

    # Model
    ccM = M.gridCC
    # conds = [1e-2]
    groundInd = ccM[:, 2] < elev
    sig = simpeg.Utils.ModelBuilder.defineBlock(
        M.gridCC, np.array([-1000, -1000, -1500]), np.array([1000, 1000, -1000]), conds
    )
    sig[~groundInd] = 1e-8
    # Set the background, not the same as the model
    sigBG = np.zeros(M.nC) + 1e-8
    sigBG[groundInd] = conds[1]

    return (M, freqs, sig, sigBG, rx_loc)


def twoLayer(conds):
    """ Returns a 2 layer model based on the conductivity values given"""
    M, freqs, rx_loc, elev = getInputs()

    # Model
    ccM = M.gridCC
    groundInd = ccM[:, 2] < elev
    botInd = ccM[:, 2] < -3000
    sig = np.zeros(M.nC) + 1e-8
    sig[groundInd] = conds[1]
    sig[botInd] = conds[0]
    # Set the background, not the same as the model
    sigBG = np.zeros(M.nC) + 1e-8
    sigBG[groundInd] = conds[1]

    return (M, freqs, sig, sigBG, rx_loc)



import scipy as sp
import numpy as np
from SimPEG.regularization import Sparse, SparseSmall, SparseDeriv, Simple
from SimPEG import Mesh, Utils


def get_2d_mesh(n_sounding, hz):
    """
        Generate 2D mesh for regularization

        hx:
        hz:

    """
    hx = np.ones(n_sounding)
    return Mesh.TensorMesh([hz, hx])


class LateralConstraint(Sparse):
    def get_grad_horizontal(
        self, xy, hz, dim=3, use_cell_weights=True, minimum_distance=None
    ):
        """
            Compute Gradient in horizontal direction using Delaunay

        """
        if use_cell_weights:
            self.cell_weights = np.tile(hz, (xy.shape[0], 1)).flatten()

        if dim == 3:
            tri = sp.spatial.Delaunay(xy)
            # Split the triangulation into connections
            edges = np.r_[
                tri.simplices[:, :2], tri.simplices[:, 1:], tri.simplices[:, [0, 2]]
            ]

            # Sort and keep uniques
            edges = np.sort(edges, axis=1)
            edges = np.unique(edges[np.argsort(edges[:, 0]), :], axis=0)
            # Compute distance
            if minimum_distance is not None:
                dx = xy[edges[:, 0], 0] - xy[edges[:, 1], 0]
                dy = xy[edges[:, 0], 1] - xy[edges[:, 1], 1]
                distance = np.sqrt(dx ** 2 + dy ** 2)
                inds = distance < minimum_distance
                edges = edges[inds, :]

            # Create 2D operator, dimensionless for now
            nN = edges.shape[0]
            nStn = xy.shape[0]
            stn, count = np.unique(edges[:, 0], return_counts=True)

            col = []
            row = []
            dm = []
            avg = []
            for ii in range(nN):
                row += [ii] * 2
                col += [edges[ii, 0], edges[ii, 1]]
                scale = count[stn == edges[ii, 0]][0]
                dm += [-1.0, 1.0]
                avg += [0.5, 0.5]

            D = sp.sparse.csr_matrix((dm, (row, col)), shape=(nN, nStn))
            A = sp.sparse.csr_matrix((avg, (row, col)), shape=(nN, nStn))

            # Kron vertically for nCz
            Grad = sp.sparse.kron(D, Utils.speye(hz.size))
            Avg = sp.sparse.kron(A, Utils.speye(hz.size))

            # Override the gradient operator in y-drection
            # This is because of ordering ... See def get_2d_mesh
            # y first then x
            self.regmesh._cellDiffyStencil = self.regmesh.cellDiffxStencil.copy()
            # Override the gradient operator in x-drection
            self.regmesh._cellDiffxStencil = Grad
            # Do the same for the averaging operator
            self.regmesh._aveCC2Fy = self.regmesh.aveCC2Fx.copy()
            self.regmesh._aveCC2Fx = Avg
            self.regmesh._aveFy2CC = self.regmesh.aveFx2CC.copy()
            self.regmesh._aveFx2CC = Avg.T
            return tri

        elif dim == 2:
            # Override the gradient operator in y-drection
            # This is because of ordering ... See def get_2d_mesh
            # y first then x
            temp_x = self.regmesh.cellDiffxStencil.copy()
            temp_y = self.regmesh.cellDiffyStencil.copy()
            self.regmesh._cellDiffyStencil = temp_x
            # Override the gradient operator in x-drection
            self.regmesh._cellDiffxStencil = temp_y
            # Do the same for the averaging operator
            temp_x = self.regmesh.aveCC2Fx.copy()
            temp_y = self.regmesh.aveCC2Fy.copy()
            self.regmesh._aveCC2Fy = temp_x
            self.regmesh._aveCC2Fx = temp_y
            temp_x = self.regmesh.aveCC2Fx.copy()
            temp_y = self.regmesh.aveCC2Fy.copy()
            self.regmesh._aveFy2CC = temp_x
            self.regmesh._aveFx2CC = temp_y
            return True



import matplotlib.pyplot as plt
import numpy as np
import numpy.lib.recfunctions as recFunc
from scipy.constants import mu_0
from scipy import interpolate as sciint

import SimPEG as simpeg
from SimPEG.EM.NSEM.SurveyNSEM import Data, Survey
from SimPEG.EM.NSEM.RxNSEM import Point_impedance1D
from SimPEG.EM.NSEM.SrcNSEM import Planewave_xy_1Dprimary
from SimPEG.EM.NSEM.Utils import MT1Danalytic, plotDataTypes as pDt


def getAppRes(NSEMdata):
    # Make impedance
    zList = []
    for src in NSEMdata.survey.srcList:
        zc = [src.freq]
        for rx in src.rxList:
            if "i" in rx.rxType:
                m = 1j
            else:
                m = 1
            zc.append(m * NSEMdata[src, rx])
        zList.append(zc)
    return [
        appResPhs(zList[i][0], np.sum(zList[i][1:3])) for i in np.arange(len(zList))
    ]


def rotateData(NSEMdata, rotAngle):
    """
    Function that rotates clockwist by rotAngle (- negative for a counter-clockwise rotation)
    """
    recData = NSEMdata.toRecArray("Complex")
    impData = rec_to_ndarr(recData[["zxx", "zxy", "zyx", "zyy"]], complex)
    # Make the rotation matrix
    # c,s,zxx,zxy,zyx,zyy = sympy.symbols('c,s,zxx,zxy,zyx,zyy')
    # rotM = sympy.Matrix([[c,-s],[s, c]])
    # zM = sympy.Matrix([[zxx,zxy],[zyx,zyy]])
    # rotM*zM*rotM.T
    # [c*(c*zxx - s*zyx) - s*(c*zxy - s*zyy), c*(c*zxy - s*zyy) + s*(c*zxx - s*zyx)],
    # [c*(c*zyx + s*zxx) - s*(c*zyy + s*zxy), c*(c*zyy + s*zxy) + s*(c*zyx + s*zxx)]])
    s = np.sin(-np.deg2rad(rotAngle))
    c = np.cos(-np.deg2rad(rotAngle))
    rotMat = np.array([[c, -s], [s, c]])
    rotData = (
        (rotMat.dot(impData.reshape(-1, 2, 2).dot(rotMat.T)))
        .transpose(1, 0, 2)
        .reshape(-1, 4)
    )
    outRec = recData.copy()
    for nr, comp in enumerate(["zxx", "zxy", "zyx", "zyy"]):
        outRec[comp] = rotData[:, nr]

    return Data.fromRecArray(outRec)


def appResPhs(freq, z):
    app_res = ((1.0 / (8e-7 * np.pi ** 2)) / freq) * np.abs(z) ** 2
    app_phs = np.arctan2(z.imag, z.real) * (180 / np.pi)
    return app_res, app_phs


def skindepth(rho, freq):
    """ Function to calculate the skindepth of EM waves"""
    return np.sqrt(rho * (1 / (freq * mu_0 * np.pi)))


def rec_to_ndarr(rec_arr, data_type=float):
    """
    Function to transform a numpy record array to a nd array.
    """
    return rec_arr.view((data_type, len(rec_arr.dtype.names)))


def makeAnalyticSolution(mesh, model, elev, freqs):

    data1D = []
    for freq in freqs:
        anaEd, anaEu, anaHd, anaHu = MT1Danalytic.getEHfields(mesh, model, freq, elev)
        anaE = anaEd + anaEu
        anaH = anaHd + anaHu

        anaZ = anaE / anaH
        # Add to the list
        data1D.append((freq, 0, 0, elev, anaZ[0]))
    dataRec = np.array(
        data1D,
        dtype=[
            ("freq", float),
            ("x", float),
            ("y", float),
            ("z", float),
            ("zyx", complex),
        ],
    )
    return dataRec


def plotMT1DModelData(problem, models, symList=None):

    # Setup the figure
    fontSize = 15

    fig = plt.figure(figsize=[9, 7])
    axM = fig.add_axes([0.075, 0.1, 0.25, 0.875])
    axM.set_xlabel("Resistivity [Ohm*m]", fontsize=fontSize)
    axM.set_xlim(1e-1, 1e5)
    axM.set_ylim(-10000, 5000)
    axM.set_ylabel("Depth [km]", fontsize=fontSize)
    axR = fig.add_axes([0.42, 0.575, 0.5, 0.4])
    axR.set_xscale("log")
    axR.set_yscale("log")
    axR.invert_xaxis()
    # axR.set_xlabel('Frequency [Hz]')
    axR.set_ylabel("Apparent resistivity [Ohm m]", fontsize=fontSize)

    axP = fig.add_axes([0.42, 0.1, 0.5, 0.4])
    axP.set_xscale("log")
    axP.invert_xaxis()
    axP.set_ylim(0, 90)
    axP.set_xlabel("Frequency [Hz]", fontsize=fontSize)
    axP.set_ylabel("Apparent phase [deg]", fontsize=fontSize)

    # if not symList:
    #   symList = ['x']*len(models)
    # Loop through the models.
    modelList = [problem.survey.mtrue]
    modelList.extend(models)
    if False:
        modelList = [problem.sigmaMap * mod for mod in modelList]
    for nr, model in enumerate(modelList):
        # Calculate the data
        if nr == 0:
            data1D = problem.dataPair(problem.survey, problem.survey.dobs).toRecArray(
                "Complex"
            )
        else:
            data1D = problem.dataPair(
                problem.survey, problem.survey.dpred(model)
            ).toRecArray("Complex")
        # Plot the data and the model
        colRat = nr / ((len(modelList) - 1.999) * 1.0)
        if colRat > 1.0:
            col = "k"
        else:
            col = plt.cm.seismic(1 - colRat)
        # The model - make the pts to plot
        meshPts = np.concatenate(
            (problem.mesh.gridN[0:1], np.kron(problem.mesh.gridN[1::], np.ones(2))[:-1])
        )
        modelPts = np.kron(1.0 / (problem.sigmaMap * model), np.ones(2,))
        axM.semilogx(modelPts, meshPts, color=col)

        ## Data
        loc = rec_to_ndarr(np.unique(data1D[["x", "y"]]))
        # Appres
        pDt.plotIsoStaImpedance(axR, loc, data1D, "zyx", "res", pColor=col)
        # Appphs
        pDt.plotIsoStaImpedance(axP, loc, data1D, "zyx", "phs", pColor=col)
        try:
            allData = np.concatenate((allData, simpeg.mkvc(data1D["zyx"], 2)), 1)
        except:
            allData = simpeg.mkvc(data1D["zyx"], 2)
    freq = simpeg.mkvc(data1D["freq"], 2)
    res, phs = appResPhs(freq, allData)

    if False:
        stdCol = "gray"
        axRtw = axR.twinx()
        axRtw.set_ylabel("Std of log10", color=stdCol)
        [(t.set_color(stdCol), t.set_rotation(-45)) for t in axRtw.get_yticklabels()]
        axPtw = axP.twinx()
        axPtw.set_ylabel("Std ", color=stdCol)
        [t.set_color(stdCol) for t in axPtw.get_yticklabels()]
        axRtw.plot(freq, np.std(np.log10(res), 1), "--", color=stdCol)
        axPtw.plot(freq, np.std(phs, 1), "--", color=stdCol)

    # Fix labels and ticks

    # yMtick = [l/1000 for l in axM.get_yticks().tolist()]
    # axM.set_yticklabels(yMtick)
    [l.set_rotation(90) for l in axM.get_yticklabels()]
    [l.set_rotation(90) for l in axR.get_yticklabels()]
    # [(t.set_color(stdCol), t.set_rotation(-45)) for t in axRtw.get_yticklabels()]
    # [t.set_color(stdCol) for t in axPtw.get_yticklabels()]
    for ax in [axM, axR, axP]:
        ax.xaxis.set_tick_params(labelsize=fontSize)
        ax.yaxis.set_tick_params(labelsize=fontSize)
    return fig


def plotImpAppRes(dataArrays, plotLoc, textStr=[]):
    """ Plots amplitude impedance and phase"""
    # Make the figure and axes
    fig, axT = plt.subplots(2, 2, sharex=True)
    axes = axT.ravel()
    fig.set_size_inches((13.5, 7.0))
    fig.suptitle(
        "{:s}\nStation at: {:.1f}x ; {:.1f}y".format(textStr, plotLoc[0], plotLoc[1])
    )
    # Have to deal with axes
    # Set log
    for ax in axes.ravel():
        ax.set_xscale("log")

    axes[0].invert_xaxis()
    axes[0].set_yscale("log")
    axes[2].set_yscale("log")
    # Set labels
    axes[2].set_xlabel("Frequency [Hz]")
    axes[3].set_xlabel("Frequency [Hz]")
    axes[0].set_ylabel("Apperent resistivity [Ohm m]")
    axes[1].set_ylabel("Apperent phase [degrees]")
    axes[1].set_ylim(-180, 180)
    axes[2].set_ylabel("Impedance amplitude [V/A]")
    axes[3].set_ylim(-180, 180)
    axes[3].set_ylabel("Impedance angle [degrees]")

    # Plot the data
    for nr, dataArray in enumerate(dataArrays):
        if nr == 1:
            parSym = "*"
        else:
            parSym = "s"
        # app res
        pDt.plotIsoStaImpedance(
            axes[0], plotLoc, dataArray, "zxy", par="res", pSym=parSym
        )
        pDt.plotIsoStaImpedance(
            axes[0], plotLoc, dataArray, "zyx", par="res", pSym=parSym
        )
        # app phs
        pDt.plotIsoStaImpedance(
            axes[1], plotLoc, dataArray, "zxy", par="phs", pSym=parSym
        )
        pDt.plotIsoStaImpedance(
            axes[1], plotLoc, dataArray, "zyx", par="phs", pSym=parSym
        )
        # imp abs
        pDt.plotIsoStaImpedance(
            axes[2], plotLoc, dataArray, "zxx", par="abs", pSym=parSym
        )
        pDt.plotIsoStaImpedance(
            axes[2], plotLoc, dataArray, "zxy", par="abs", pSym=parSym
        )
        pDt.plotIsoStaImpedance(
            axes[2], plotLoc, dataArray, "zyx", par="abs", pSym=parSym
        )
        pDt.plotIsoStaImpedance(
            axes[2], plotLoc, dataArray, "zyy", par="abs", pSym=parSym
        )
        # imp abs
        pDt.plotIsoStaImpedance(
            axes[3], plotLoc, dataArray, "zxx", par="phs", pSym=parSym
        )
        pDt.plotIsoStaImpedance(
            axes[3], plotLoc, dataArray, "zxy", par="phs", pSym=parSym
        )
        pDt.plotIsoStaImpedance(
            axes[3], plotLoc, dataArray, "zyx", par="phs", pSym=parSym
        )
        pDt.plotIsoStaImpedance(
            axes[3], plotLoc, dataArray, "zyy", par="phs", pSym=parSym
        )

    return fig, axes


def printTime():
    import time

    print(time.strftime("%a, %d %b %Y %H:%M:%S +0000", time.localtime()))


def convert3Dto1Dobject(NSEMdata, rxType3D="yx"):

    # Find the unique locations
    # Need to find the locations
    recDataTemp = NSEMdata.toRecArray().data.flatten()
    # Check if survey.std has been assigned.
    ## NEED TO: write this...
    # Calculte and add the DET of the tensor to the recArray
    if "det" in rxType3D:
        Zon = (recDataTemp["zxxr"] + 1j * recDataTemp["zxxi"]) * (
            recDataTemp["zyyr"] + 1j * recDataTemp["zyyi"]
        )
        Zoff = (recDataTemp["zxyr"] + 1j * recDataTemp["zxyi"]) * (
            recDataTemp["zyxr"] + 1j * recDataTemp["zyxi"]
        )
        det = np.sqrt(Zon - Zoff)
        recData = recFunc.append_fields(
            recDataTemp, ["zdetr", "zdeti"], [det.real, det.imag]
        )
    else:
        recData = recDataTemp

    uniLocs = rec_to_ndarr(np.unique(recData[["x", "y", "z"]]))
    mtData1DList = []
    if "xy" in rxType3D:
        corr = -1  # Shift the data to comply with the quadtrature of the 1d problem
    else:
        corr = 1
    for loc in uniLocs:
        # Make the receiver list
        rx1DList = []
        rx1DList.append(Point_impedance1D(simpeg.mkvc(loc, 2).T, "real"))
        rx1DList.append(Point_impedance1D(simpeg.mkvc(loc, 2).T, "imag"))
        # Source list
        locrecData = recData[
            np.sqrt(np.sum((rec_to_ndarr(recData[["x", "y", "z"]]) - loc) ** 2, axis=1))
            < 1e-5
        ]
        dat1DList = []
        src1DList = []
        for freq in locrecData["freq"]:
            src1DList.append(Planewave_xy_1Dprimary(rx1DList, freq))
            for comp in ["r", "i"]:
                dat1DList.append(
                    corr * locrecData[rxType3D + comp][locrecData["freq"] == freq]
                )

        # Make the survey
        sur1D = Survey(src1DList)

        # Make the data
        dataVec = np.hstack(dat1DList)
        dat1D = Data(sur1D, dataVec)
        sur1D.dobs = dataVec
        # Need to take NSEMdata.survey.std and split it as well.
        std = 0.05
        sur1D.std = np.abs(sur1D.dobs * std)  # + 0.01*np.linalg.norm(sur1D.dobs)
        mtData1DList.append(dat1D)

    # Return the the list of data.
    return mtData1DList


def resampleNSEMdataAtFreq(NSEMdata, freqs):
    """
    Function to resample NSEMdata at set of frequencies

    """

    # Make a rec array
    NSEMrec = NSEMdata.toRecArray().data

    # Find unique locations
    uniLoc = np.unique(NSEMrec[["x", "y", "z"]])
    uniFreq = NSEMdata.survey.freqs
    # Get the comps
    dNames = NSEMrec.dtype

    # Loop over all the locations and interpolate
    for loc in uniLoc:
        # Find the index of the station
        ind = (
            np.sqrt(
                np.sum(
                    (rec_to_ndarr(NSEMrec[["x", "y", "z"]]) - rec_to_ndarr(loc)) ** 2,
                    axis=1,
                )
            )
            < 1.0
        )  # Find dist of 1 m accuracy
        # Make a temporary recArray and interpolate all the components
        tArrRec = np.concatenate(
            (
                simpeg.mkvc(freqs, 2),
                np.ones((len(freqs), 1)) * rec_to_ndarr(loc),
                np.nan * np.ones((len(freqs), 12)),
            ),
            axis=1,
        ).view(dNames)
        for comp in [
            "zxxr",
            "zxxi",
            "zxyr",
            "zxyi",
            "zyxr",
            "zyxi",
            "zyyr",
            "zyyi",
            "tzxr",
            "tzxi",
            "tzyr",
            "tzyi",
        ]:
            int1d = sciint.interp1d(
                NSEMrec[ind]["freq"], NSEMrec[ind][comp], bounds_error=False
            )
            tArrRec[comp] = simpeg.mkvc(int1d(freqs), 2)

        # Join together
        try:
            outRecArr = recFunc.stack_arrays((outRecArr, tArrRec))
        except NameError:
            outRecArr = tArrRec

    # Make the NSEMdata and return
    return Data.fromRecArray(outRecArr)



#  Copyright (c) 2021 Mira Geoscience Ltd.
#
#  This file is part of geoapps.
#
#  geoapps is distributed under the terms and conditions of the MIT License
#  (see LICENSE file at the root of this source code package).

import discretize
import numpy as np
from geoh5py.objects import BlockModel, ObjectBase
from geoh5py.workspace import Workspace
from ipywidgets import (
    Dropdown,
    FloatText,
    HBox,
    Label,
    RadioButtons,
    Text,
    ToggleButton,
    VBox,
)
from scipy.interpolate import LinearNDInterpolator
from scipy.spatial import cKDTree

from geoapps.selection import ObjectDataSelection, TopographyOptions
from geoapps.utils import weighted_average


class DataInterpolation(ObjectDataSelection):
    """
    Transfer data from one object to another, or onto a 3D BlockModel
    """

    defaults = {
        "select_multiple": True,
        "h5file": "../../assets/FlinFlon.geoh5",
        "objects": "Inversion_VTEM_Model",
        "data": ["Iteration_7_model"],
        "core_cell_size": "50, 50, 50",
        "depth_core": 500,
        "expansion_fact": 1.05,
        "max_distance": 2e3,
        "max_depth": 1e3,
        "method": "Inverse Distance",
        "new_grid": "InterpGrid",
        "no_data_value": 1e-8,
        "out_mode": "To Object",
        "out_object": "O2O_Interp_25m",
        "padding_distance": "0, 0, 0, 0, 0, 0",
        "skew_angle": 0,
        "skew_factor": 1.0,
        "space": "Log",
        "topography": {"objects": "Topography", "data": "Z"},
    }

    def __init__(self, use_defaults=True, **kwargs):

        if use_defaults:
            kwargs = self.apply_defaults(**kwargs)

        self._core_cell_size = Text(
            description="Smallest cells",
        )
        self._depth_core = FloatText(
            description="Core depth (m)",
        )
        self._expansion_fact = FloatText(
            description="Expansion factor",
        )
        self._max_distance = FloatText(
            description="Maximum distance (m)",
        )
        self._max_depth = FloatText(
            description="Maximum depth (m)",
        )
        self._method = RadioButtons(
            options=["Nearest", "Linear", "Inverse Distance"],
        )
        self._new_grid = Text(
            description="Name",
        )
        self._no_data_value = FloatText()
        self._out_mode = RadioButtons(
            options=["To Object", "Create 3D Grid"],
        )
        self._out_object = Dropdown()
        self._padding_distance = Text(
            description="Pad Distance (W, E, N, S, D, U)",
        )
        self._skew_angle = FloatText(
            description="Azimuth (d.dd)",
        )
        self._skew_factor = FloatText(
            description="Factor (>0)",
        )
        self._space = RadioButtons(options=["Linear", "Log"])
        self._xy_extent = Dropdown(
            description="Object hull",
        )
        self._xy_reference = Dropdown(
            description="Lateral Extent", style={"description_width": "initial"}
        )

        def object_pick(_):
            self.object_pick()

        self.objects.observe(object_pick, names="value")

        self.method_skew = VBox(
            [Label("Skew parameters"), self.skew_angle, self.skew_factor]
        )
        self.method_panel = VBox([self.method])
        self.destination_panel = VBox([self.out_mode, self.out_object])
        self.new_grid_panel = VBox(
            [
                self.new_grid,
                self.xy_reference,
                self.core_cell_size,
                self.depth_core,
                self.padding_distance,
                self.expansion_fact,
            ]
        )

        self.method.observe(self.method_update)
        self.out_mode.observe(self.out_update)

        super().__init__(**kwargs)

        def interpolate_call(_):
            self.interpolate_call()
            self.update_objects_choices()

        self.trigger.on_click(interpolate_call)
        self.trigger.description = "Interpolate"
        self._topography = TopographyOptions()
        self.topography.offset.disabled = True
        self.topography.options.options = ["None", "Object", "Constant"]

        if getattr(self, "_workspace", None) is not None:
            self.topography.workspace = self._workspace
            if "topography" in kwargs.keys():
                self.topography.__populate__(**kwargs["topography"])

        self.parameter_choices = Dropdown(
            description="Interpolation Parameters",
            options=[
                "Method",
                "Scaling",
                "Horizontal Extent",
                "Vertical Extent",
                "No-data-value",
            ],
            style={"description_width": "initial"},
        )
        self.parameters = {
            "Method": self.method_panel,
            "Scaling": self.space,
            "Horizontal Extent": VBox(
                [
                    self.max_distance,
                    self.xy_extent,
                ]
            ),
            "Vertical Extent": VBox([self.topography.main, self.max_depth]),
            "No-data-value": self.no_data_value,
        }

        self.parameter_panel = HBox([self.parameter_choices, self.method_panel])
        self.ga_group_name.description = "Output Label:"
        self.ga_group_name.value = "_Interp"
        self.parameter_choices.observe(self.parameter_change)
        self._main = VBox(
            [
                self.project_panel,
                HBox(
                    [
                        VBox([Label("Source"), self.main]),
                        VBox([Label("Destination"), self.destination_panel]),
                    ]
                ),
                self.parameter_panel,
                self.output_panel,
            ]
        )

    @property
    def core_cell_size(self):
        """
        :obj:`ipywidgets.Text()`
        """
        return self._core_cell_size

    @property
    def depth_core(self):
        """
        :obj:`ipywidgets.FloatText()`
        """
        return self._depth_core

    @property
    def expansion_fact(self):
        """
        :obj:`ipywidgets.FloatText()`
        """
        return self._expansion_fact

    @property
    def max_distance(self):
        """
        :obj:`ipywidgets.FloatText()`
        """
        return self._max_distance

    @property
    def max_depth(self):
        """
        :obj:`ipywidgets.FloatText()`
        """
        return self._max_depth

    @property
    def method(self):
        """
        :obj:`ipywidgets.RadioButtons()`
        """
        return self._method

    @property
    def new_grid(self):
        """
        :obj:`ipywidgets.Text()`
        """
        return self._new_grid

    @property
    def no_data_value(self):
        """
        :obj:`ipywidgets.FloatText()`
        """
        return self._no_data_value

    @property
    def out_mode(self):
        """
        :obj:`ipywidgets.RadioButtons()`
        """
        return self._out_mode

    @property
    def out_object(self):
        """
        :obj:`ipywidgets.Dropdown()`
        """
        return self._out_object

    @property
    def padding_distance(self):
        """
        :obj:`ipywidgets.Text()`
        """
        return self._padding_distance

    @property
    def skew_angle(self):
        """
        :obj:`ipywidgets.FloatText()`
        """
        return self._skew_angle

    @property
    def skew_factor(self):
        """
        :obj:`ipywidgets.FloatText()`
        """
        return self._skew_factor

    @property
    def space(self):
        """
        :obj:`ipywidgets.RadioButtons()`
        """
        return self._space

    @property
    def topography(self):
        """
        :obj:`geoapps.TopographyOptions()`
        """
        return self._topography

    @property
    def xy_extent(self):
        """
        :obj:`ipywidgets.Dropdown()`
        """
        return self._xy_extent

    @property
    def xy_reference(self):
        """
        :obj:`ipywidgets.Dropdown()`
        """
        return self._xy_reference

    @property
    def workspace(self):
        """
        Target geoh5py workspace
        """
        if (
            getattr(self, "_workspace", None) is None
            and getattr(self, "_h5file", None) is not None
        ):
            self.workspace = Workspace(self.h5file)
        return self._workspace

    @workspace.setter
    def workspace(self, workspace):

        assert isinstance(workspace, Workspace), f"Workspace must of class {Workspace}"
        self._workspace = workspace
        self._h5file = workspace.h5file

        self.update_objects_choices()

        if getattr(self, "topography", None) is not None:
            self.topography.workspace = workspace

    def parameter_change(self, _):
        self.parameter_panel.children = [
            self.parameter_choices,
            self.parameters[self.parameter_choices.value],
        ]

    def update_objects_choices(self):
        # Refresh the list of objects for all
        self.update_objects_list()

        value = self.out_object.value
        self.out_object.options = self.objects.options
        if value in self.out_object.options:
            self.out_object.value = value

        value = self.xy_reference.value
        self.xy_reference.options = self.objects.options
        if value in self.xy_reference.options:
            self.xy_reference.value = value

        value = self.xy_extent.value
        self.xy_extent.options = self.objects.options
        if value in self.xy_extent.options:
            self.xy_extent.value = value

    def method_update(self, _):
        if self.method.value == "Inverse Distance":
            self.method_panel.children = [self.method, self.method_skew]
        elif self.method.value == "Linear":
            self.method_panel.children = [
                self.method,
                Label("Warning! Very slow on 3D objects"),
            ]
        else:
            self.method_panel.children = [self.method]

    def out_update(self, _):
        if self.out_mode.value == "To Object":
            self.destination_panel.children = [self.out_mode, self.out_object]
        else:
            self.destination_panel.children = [self.out_mode, self.new_grid_panel]

    def interpolate_call(self):

        for entity in self._workspace.get_entity(self.objects.value):
            if isinstance(entity, ObjectBase):
                object_from = entity

        if hasattr(object_from, "centroids"):
            xyz = object_from.centroids.copy()
        elif hasattr(object_from, "vertices"):
            xyz = object_from.vertices.copy()
        else:
            return

        if len(self.data.value) == 0:
            print("No data selected")
            return
        # Create a tree for the input mesh
        tree = cKDTree(xyz)

        if self.out_mode.value == "To Object":

            for entity in self._workspace.get_entity(self.out_object.value):
                if isinstance(entity, ObjectBase):
                    self.object_out = entity

            if hasattr(self.object_out, "centroids"):
                xyz_out = self.object_out.centroids.copy()
            elif hasattr(self.object_out, "vertices"):
                xyz_out = self.object_out.vertices.copy()

        else:

            for entity in self._workspace.get_entity(self.xy_reference.value):
                if isinstance(entity, ObjectBase):
                    ref_in = entity

            if hasattr(ref_in, "centroids"):
                xyz_ref = ref_in.centroids
            elif hasattr(ref_in, "vertices"):
                xyz_ref = ref_in.vertices

            # Find extent of grid
            h = np.asarray(self.core_cell_size.value.split(",")).astype(float).tolist()

            pads = (
                np.asarray(self.padding_distance.value.split(","))
                .astype(float)
                .tolist()
            )

            # Use discretize to build a tensor mesh
            delta_z = xyz_ref[:, 2].max() - xyz_ref[:, 2]
            xyz_ref[delta_z > self.depth_core.value, 2] = (
                xyz_ref[:, 2].max() - self.depth_core.value
            )
            depth_core = (
                self.depth_core.value
                - (xyz_ref[:, 2].max() - xyz_ref[:, 2].min())
                + h[2]
            )
            mesh = discretize.utils.meshutils.mesh_builder_xyz(
                xyz_ref,
                h,
                padding_distance=[
                    [pads[0], pads[1]],
                    [pads[2], pads[3]],
                    [pads[4], pads[5]],
                ],
                depth_core=depth_core,
                expansion_factor=self.expansion_fact.value,
            )
            self.object_out = BlockModel.create(
                self.workspace,
                origin=[mesh.x0[0], mesh.x0[1], xyz_ref[:, 2].max()],
                u_cell_delimiters=mesh.vectorNx - mesh.x0[0],
                v_cell_delimiters=mesh.vectorNy - mesh.x0[1],
                z_cell_delimiters=-(mesh.x0[2] + mesh.hz.sum() - mesh.vectorNz[::-1]),
                name=self.new_grid.value,
            )

            # Try to recenter on nearest
            # Find nearest cells
            rad, ind = tree.query(self.object_out.centroids)
            ind_nn = np.argmin(rad)

            d_xyz = self.object_out.centroids[ind_nn, :] - xyz[ind[ind_nn], :]

            self.object_out.origin = np.r_[self.object_out.origin.tolist()] - d_xyz

            xyz_out = self.object_out.centroids.copy()

        xyz_out_orig = xyz_out.copy()

        values, sign, dtype = {}, {}, {}
        for field in self.data.value:
            model_in = object_from.get_data(field)[0]
            values[field] = np.asarray(model_in.values, dtype=float).copy()
            dtype[field] = model_in.values.dtype
            values[field][values[field] == self.no_data_value.value] = np.nan
            if self.space.value == "Log":
                sign[field] = np.sign(values[field])
                values[field] = np.log(np.abs(values[field]))
            else:
                sign[field] = np.ones_like(values[field])

        values_interp = {}
        rad, ind = tree.query(xyz_out)
        if self.method.value == "Linear":

            for key, value in values.items():
                F = LinearNDInterpolator(xyz, value)
                values_interp[key] = F(xyz_out)

        elif self.method.value == "Inverse Distance":

            angle = np.deg2rad((450.0 - np.asarray(self.skew_angle.value)) % 360.0)
            rotation = np.r_[
                np.c_[np.cos(angle), np.sin(angle)],
                np.c_[-np.sin(angle), np.cos(angle)],
            ]
            center = np.mean(xyz, axis=0).reshape((3, 1))
            xyz -= np.kron(center, np.ones(xyz.shape[0])).T
            xyz[:, :2] = np.dot(rotation, xyz[:, :2].T).T
            xyz[:, 1] *= self.skew_factor.value
            xyz_out -= np.kron(center, np.ones(xyz_out.shape[0])).T
            xyz_out[:, :2] = np.dot(rotation, xyz_out[:, :2].T).T
            xyz_out[:, 1] *= self.skew_factor.value
            vals, ind_inv = weighted_average(
                xyz,
                xyz_out,
                list(values.values()),
                threshold=1e-1,
                n=8,
                return_indices=True,
            )

            for key, val in zip(list(values.keys()), vals):
                values_interp[key] = val
                sign[key] = sign[key][ind_inv[:, 0]]

        else:
            # Find nearest cells
            for key, value in values.items():

                values_interp[key] = value[ind]
                sign[key] = sign[key][ind]

        for key in values_interp.keys():
            if self.space.value == "Log":
                values_interp[key] = sign[key] * np.exp(values_interp[key])

            values_interp[key][np.isnan(values_interp[key])] = self.no_data_value.value
            values_interp[key][rad > self.max_distance.value] = self.no_data_value.value

        # if hasattr(self.object_out, "centroids"):
        #     xyz_out = self.object_out.centroids
        # elif hasattr(self.object_out, "vertices"):
        #     xyz_out = self.object_out.vertices

        top = np.zeros(xyz_out.shape[0], dtype="bool")
        bottom = np.zeros(xyz_out.shape[0], dtype="bool")
        if self.topography.options.value == "Object" and self.workspace.get_entity(
            self.topography.objects.value
        ):

            for entity in self._workspace.get_entity(self.topography.objects.value):
                if isinstance(entity, ObjectBase):
                    topo_obj = entity

            if getattr(topo_obj, "vertices", None) is not None:
                topo = topo_obj.vertices
            else:
                topo = topo_obj.centroids

            if self.topography.data.value != "Z":
                topo[:, 2] = topo_obj.get_data(self.topography.data.value)[0].values

            lin_interp = LinearNDInterpolator(topo[:, :2], topo[:, 2])
            z_interp = lin_interp(xyz_out_orig[:, :2])

            ind_nan = np.isnan(z_interp)
            if any(ind_nan):
                tree = cKDTree(topo[:, :2])
                _, ind = tree.query(xyz_out_orig[ind_nan, :2])
                z_interp[ind_nan] = topo[ind, 2]

            top = xyz_out_orig[:, 2] > z_interp
            if self.max_depth.value is not None:
                bottom = np.abs(xyz_out_orig[:, 2] - z_interp) > self.max_depth.value

        elif (
            self.topography.options.value == "Constant"
            and self.topography.constant.value is not None
        ):
            top = xyz_out_orig[:, 2] > self.topography.constant.value
            if self.max_depth.value is not None:
                bottom = (
                    np.abs(xyz_out_orig[:, 2] - self.topography.constant.value)
                    > self.max_depth.value
                )

        for key in values_interp.keys():
            values_interp[key][top] = self.no_data_value.value
            values_interp[key][bottom] = self.no_data_value.value

        if self.xy_extent.value is not None and self.workspace.get_entity(
            self.xy_extent.value
        ):

            for entity in self._workspace.get_entity(self.xy_extent.value):
                if isinstance(entity, ObjectBase):
                    xy_ref = entity
            if hasattr(xy_ref, "centroids"):
                xy_ref = xy_ref.centroids
            elif hasattr(xy_ref, "vertices"):
                xy_ref = xy_ref.vertices

            tree = cKDTree(xy_ref[:, :2])
            rad, _ = tree.query(xyz_out_orig[:, :2])
            for key in values_interp.keys():
                values_interp[key][
                    rad > self.max_distance.value
                ] = self.no_data_value.value

        for key in values_interp.keys():
            if dtype[field] == np.dtype("int32"):
                primitive = "integer"
                vals = np.round(values_interp[key]).astype(dtype[field])
            else:
                primitive = "float"
                vals = values_interp[key].astype(dtype[field])

            self.object_out.add_data(
                {key + self.ga_group_name.value: {"values": vals, "type": primitive}}
            )

        if self.live_link.value:
            self.live_link_output(self.object_out)

        self.workspace.finalize()

    def object_pick(self):
        if self.objects.value in self.xy_reference.options:
            self.xy_reference.value = self.objects.value



import numpy as np
import scipy.sparse as sp

from .. import Survey
from .. import Utils


class BaseRichardsRx(Survey.BaseTimeRx):
    """Richards Receiver Object"""

    def __init__(self, locs, times):
        self.locs = locs
        self.times = times
        self._Ps = {}


class PressureRx(BaseRichardsRx):
    """Richards Receiver Object"""

    def __call__(self, U, prob):
        P = self.getP(prob.mesh, prob.timeMesh)
        u = np.concatenate(U)
        return P * u

    def deriv(self, U, prob, du_dm_v=None, v=None, adjoint=False):
        P = self.getP(prob.mesh, prob.timeMesh)
        if not adjoint:
            return P * du_dm_v  # + 0 for dRx_dm contribution
        assert v is not None, "v must be provided if computing adjoint"
        return P.T * v, Utils.Zero()


class SaturationRx(BaseRichardsRx):
    """Richards Receiver Object"""

    def __call__(self, U, prob):
        # The water retention curve model should have been updated in the prob
        P = self.getP(prob.mesh, prob.timeMesh)
        usat = np.concatenate([prob.water_retention(ui) for ui in U])
        return P * usat

    def deriv(self, U, prob, du_dm_v=None, v=None, adjoint=False):
        # The water retention curve model should have been updated in the prob

        P = self.getP(prob.mesh, prob.timeMesh)
        dT_du = sp.block_diag([prob.water_retention.derivU(ui) for ui in U])

        if prob.water_retention.needs_model:
            dT_dm = sp.vstack([prob.water_retention.derivM(ui) for ui in U])
        else:
            dT_dm = Utils.Zero()

        if v is None and not adjoint:
            # this is called by the fullJ in the problem
            return P * (dT_du * du_dm_v) + P * dT_dm
        if not adjoint:
            return P * (dT_du * du_dm_v) + P * (dT_dm * v)

        # for the adjoint return both parts of the sum separately
        assert v is not None, "v must be provided if computing adjoint"
        PTv = P.T * v
        return dT_du.T * PTv, dT_dm.T * PTv


class RichardsSurvey(Survey.BaseSurvey):
    """RichardsSurvey"""

    rxList = None

    def __init__(self, rxList, **kwargs):
        self.rxList = rxList
        Survey.BaseSurvey.__init__(self, **kwargs)

    @property
    def nD(self):
        return np.array([rx.nD for rx in self.rxList]).sum()

    @Utils.count
    @Utils.requires("prob")
    def dpred(self, m, f=None):
        """Create the projected data from a model.
        The field, f, (if provided) will be used for the predicted data
        instead of recalculating the fields (which may be expensive!).

        .. math::
            d_\\text{pred} = P(f(m), m)

        Where P is a projection of the fields onto the data space.
        """
        if f is None:
            f = self.prob.fields(m)

        return Utils.mkvc(self(f))

    @Utils.requires("prob")
    def __call__(self, U):
        Ds = list(range(len(self.rxList)))

        for ii, rx in enumerate(self.rxList):
            Ds[ii] = rx(U, self.prob)

        return np.concatenate(Ds)

    @Utils.requires("prob")
    def deriv(self, U, du_dm_v=None, v=None):
        """The Derivative with respect to the model."""
        dd_dm = [
            rx.deriv(U, self.prob, du_dm_v=du_dm_v, v=v, adjoint=False)
            for rx in self.rxList
        ]
        return np.concatenate(dd_dm)

    @Utils.requires("prob")
    def derivAdjoint(self, U, v=None):
        """The adjoint derivative with respect to the model."""
        dd_du = list(range(len(self.rxList)))
        dd_dm = list(range(len(self.rxList)))
        cnt = 0
        for ii, rx in enumerate(self.rxList):
            dd_du[ii], dd_dm[ii] = rx.deriv(
                U, self.prob, v=v[cnt : cnt + rx.nD], adjoint=True
            )
            cnt += rx.nD
        return np.sum(dd_du, axis=0), np.sum(dd_dm, axis=0)



#  Copyright (c) 2021 Mira Geoscience Ltd.
#
#  This file is part of geoapps.
#
#  geoapps is distributed under the terms and conditions of the MIT License
#  (see LICENSE file at the root of this source code package).

import os

import numpy as np
import plotly.express as px
import plotly.graph_objects as go
from ipywidgets import (
    Checkbox,
    Dropdown,
    FloatText,
    HBox,
    IntSlider,
    Label,
    Layout,
    ToggleButton,
    VBox,
    interactive_output,
)

from geoapps.plotting import format_axis, normalize
from geoapps.selection import ObjectDataSelection
from geoapps.utils import random_sampling, symlog


class ScatterPlots(ObjectDataSelection):
    """
    Application for 2D and 3D crossplots of data using symlog scaling
    """

    defaults = {
        "h5file": "../../assets/FlinFlon.geoh5",
        "objects": "geochem",
        "data": ["Al2O3", "CaO", "V", "MgO", "Ba"],
        "x": "Al2O3",
        "y": "CaO",
        "z": "Ba",
        "y_log": True,
        "z_log": True,
        "z_active": True,
        "color": "V",
        "color_active": True,
        "color_log": True,
        "size": "MgO",
        "size_active": True,
        "color_maps": "inferno",
        "refresh": True,
        "refresh_trigger": True,
    }

    def __init__(self, static=False, **kwargs):
        self.static = static
        self.select_multiple = True
        self._add_groups = True
        self.custom_colormap = []
        self._indices = None

        def channel_bounds_setter(caller):
            self.set_channel_bounds(caller["owner"].name)

        self._downsampling = IntSlider(
            description="Population Downsampling:",
            min=1,
            style={"description_width": "initial"},
            continuous_update=False,
        )

        self._x = Dropdown(description="Data:")
        self._x.observe(channel_bounds_setter, names="value")
        self._x.name = "x"
        self._x_active = Checkbox(description="Active", value=True, indent=False)
        self._x_log = Checkbox(
            description="Log10",
            value=False,
            indent=False,
        )
        self._x_thresh = FloatText(
            description="Threshold",
            value=1e-1,
        )
        self._x_min = FloatText(
            description="Min",
        )
        self._x_max = FloatText(
            description="Max",
        )
        self._x_panel = VBox(
            [
                self._x_active,
                HBox([self._x]),
                HBox([self._x_log, self._x_thresh]),
                HBox([self._x_min, self._x_max]),
            ]
        )

        self._y = Dropdown(description="Data:")
        self._y.observe(channel_bounds_setter, names="value")
        self._y.name = "y"
        self._y_active = Checkbox(
            description="Active",
            value=True,
            indent=False,
        )
        self._y_log = Checkbox(description="Log10", value=False, indent=False)
        self._y_thresh = FloatText(
            description="Threshold",
            value=1e-1,
        )
        self._y_min = FloatText(
            description="Min",
        )
        self._y_max = FloatText(
            description="Max",
        )
        self._y_panel = VBox(
            [
                self._y_active,
                HBox([self._y]),
                HBox([self._y_log, self._y_thresh]),
                HBox([self._y_min, self._y_max]),
            ]
        )

        self._z = Dropdown(description="Data:")
        self._z.observe(channel_bounds_setter, names="value")
        self._z.name = "z"
        self._z_active = Checkbox(
            description="Active",
            value=False,
            indent=False,
        )
        self._z_log = Checkbox(
            description="Log10",
            value=False,
            indent=False,
        )
        self._z_thresh = FloatText(
            description="Threshold",
            value=1e-1,
        )
        self._z_min = FloatText(
            description="Min",
        )
        self._z_max = FloatText(
            description="Max",
        )
        self._z_panel = VBox(
            [
                self._z_active,
                HBox([self._z]),
                HBox([self._z_log, self._z_thresh]),
                HBox([self._z_min, self._z_max]),
            ]
        )
        self._color = Dropdown(description="Data:")
        self._color.observe(channel_bounds_setter, names="value")
        self._color.name = "color"
        self._color_log = Checkbox(
            description="Log10",
            value=False,
            indent=False,
        )
        self._color_thresh = FloatText(
            description="Threshold",
            value=1e-1,
        )
        self._color_active = Checkbox(
            description="Active",
            value=False,
            indent=False,
        )
        self._color_maps = Dropdown(
            description="Colormaps",
            options=px.colors.named_colorscales(),
            value="viridis",
        )
        self._color_min = FloatText(
            description="Min",
        )
        self._color_max = FloatText(
            description="Max",
        )
        self._color_panel = VBox(
            [
                self._color_active,
                HBox([self._color]),
                self._color_maps,
                HBox([self._color_log, self._color_thresh]),
                HBox([self._color_min, self._color_max]),
            ]
        )

        self._size = Dropdown(description="Data:")
        self._size.observe(channel_bounds_setter, names="value")
        self._size.name = "size"
        self._size_active = Checkbox(
            description="Active",
            value=False,
            indent=False,
        )
        self._size_log = Checkbox(
            description="Log10",
            value=False,
            indent=False,
        )
        self._size_thresh = FloatText(
            description="Threshold",
            value=1e-1,
        )
        self._size_markers = IntSlider(
            min=1, max=100, value=20, description="Marker size", continuous_update=False
        )
        self._size_min = FloatText(
            description="Min",
        )
        self._size_max = FloatText(
            description="Max",
        )

        self._size_panel = VBox(
            [
                self._size_active,
                HBox([self._size]),
                self._size_markers,
                HBox([self._size_log, self._size_thresh]),
                HBox([self._size_min, self._size_max]),
            ]
        )
        self._refresh_trigger = ToggleButton(description="Refresh Plot", value=False)

        # Wrap all axis panels into dropdown
        def axes_pannels_trigger(_):
            self.axes_pannels_trigger()

        self.panels = {
            "X-axis": self._x_panel,
            "Y-axis": self._y_panel,
            "Z-axis": self._z_panel,
            "Color": self._color_panel,
            "Size": self._size_panel,
        }
        self.axes_pannels = Dropdown(
            options=["X-axis", "Y-axis", "Z-axis", "Color", "Size"],
            layout=Layout(width="300px"),
        )
        self.axes_pannels.observe(axes_pannels_trigger, names="value")
        self.axes_options = VBox([self.axes_pannels, self._x_panel])
        self.data_channels = {}
        self.data.observe(self.update_choices, names="value")
        self.objects.observe(self.update_objects, names="value")
        self.downsampling.observe(self.update_downsampling, names="value")

        super().__init__(**self.apply_defaults(**kwargs))

        self.figure = go.FigureWidget()

        self.crossplot = interactive_output(
            self.plot_selection,
            {
                "x": self.x,
                "x_log": self.x_log,
                "x_active": self.x_active,
                "x_thresh": self.x_thresh,
                "x_min": self.x_min,
                "x_max": self.x_max,
                "y": self.y,
                "y_log": self.y_log,
                "y_thresh": self.y_thresh,
                "y_active": self.y_active,
                "y_min": self.y_min,
                "y_max": self.y_max,
                "z": self.z,
                "z_log": self.z_log,
                "z_thresh": self.z_thresh,
                "z_active": self.z_active,
                "z_min": self.z_min,
                "z_max": self.z_max,
                "color": self.color,
                "color_log": self.color_log,
                "color_thresh": self.color_thresh,
                "color_active": self.color_active,
                "color_maps": self.color_maps,
                "color_min": self.color_min,
                "color_max": self.color_max,
                "size": self.size,
                "size_log": self.size_log,
                "size_thresh": self.size_thresh,
                "size_active": self.size_active,
                "size_markers": self.size_markers,
                "size_min": self.size_min,
                "size_max": self.size_max,
                "refresh_trigger": self.refresh_trigger,
            },
        )

        def write_html(_):
            self.write_html()

        self.trigger.on_click(write_html)
        self.trigger.description = "Save HTML"

        if self.static:
            self._main = VBox(
                [
                    self.project_panel,
                    HBox([self.objects, self.data]),
                    self.downsampling,
                    self.axes_options,
                    self.trigger,
                ]
            )
        else:
            self._main = VBox(
                [
                    self.project_panel,
                    HBox([self.objects, self.data]),
                    VBox([Label("Downsampling"), self.downsampling]),
                    self.axes_options,
                    self.trigger,
                    self.figure,
                ]
            )

    @property
    def n_values(self):
        """
        Number of values contained by the current object
        """

        obj, _ = self.get_selected_entities()
        if obj is not None:
            # Check number of points
            if hasattr(obj, "centroids"):
                return obj.n_cells
            elif hasattr(obj, "vertices"):
                return obj.n_vertices
        return None

    @property
    def indices(self):
        """
        Bool or array of int
        Indices of data to be plotted
        """
        if getattr(self, "_indices", None) is None:
            if self.n_values is not None:
                self._indices = np.arange(self.n_values)
            else:
                return None

        return self._indices

    @property
    def color(self):
        """
        :obj:`ipywidgets.Dropdown`
        """
        return self._color

    @property
    def color_log(self):
        """
        :obj:`ipywidgets.Checkbox`
        """
        return self._color_log

    @property
    def color_thresh(self):
        """
        :obj:`ipywidgets.FloatText`
        """
        return self._color_thresh

    @property
    def color_active(self):
        """
        :obj:`ipywidgets.Checkbox`
        """
        return self._color_active

    @property
    def color_maps(self):
        """
        :obj:`ipywidgets.Dropdown`
        """
        return self._color_maps

    @property
    def color_min(self):
        """
        :obj:`ipywidgets.Text`
        """
        return self._color_min

    @property
    def color_max(self):
        """
        :obj:`ipywidgets.Text`
        """
        return self._color_max

    @property
    def downsampling(self):
        """
        :obj:`ipywidgets.IntSlider`
        """
        return self._downsampling

    @property
    def refresh_trigger(self):
        """"""
        return self._refresh_trigger

    @property
    def size(self):
        """
        :obj:`ipywidgets.Dropdown`
        """
        return self._size

    @property
    def size_active(self):
        """
        :obj:`ipywidgets.Checkbox`
        """
        return self._size_active

    @property
    def size_log(self):
        """
        :obj:`ipywidgets.Checkbox`
        """
        return self._size_log

    @property
    def size_thresh(self):
        """
        :obj:`ipywidgets.FloatText`
        """
        return self._size_thresh

    @property
    def size_markers(self):
        """
        :obj:`ipywidgets.IntSlider`
        """
        return self._size_markers

    @property
    def size_min(self):
        """
        :obj:`ipywidgets.Text`
        """
        return self._size_min

    @property
    def size_max(self):
        """
        :obj:`ipywidgets.Text`
        """
        return self._size_max

    @property
    def x(self):
        """
        :obj:`ipywidgets.Dropdown`
        """
        return self._x

    @property
    def x_active(self):
        """
        :obj:`ipywidgets.Checkbox`
        """
        return self._x_active

    @property
    def x_log(self):
        """
        :obj:`ipywidgets.Checkbox`
        """
        return self._x_log

    @property
    def x_thresh(self):
        """
        :obj:`ipywidgets.FloatText`
        """
        return self._x_thresh

    @property
    def x_min(self):
        """
        :obj:`ipywidgets.Text`
        """
        return self._x_min

    @property
    def x_max(self):
        """
        :obj:`ipywidgets.Text`
        """
        return self._x_max

    @property
    def y(self):
        """
        :obj:`ipywidgets.Dropdown`
        """
        return self._y

    @property
    def y_active(self):
        """
        :obj:`ipywidgets.Checkbox`
        """
        return self._y_active

    @property
    def y_log(self):
        """
        :obj:`ipywidgets.Checkbox`
        """
        return self._y_log

    @property
    def y_thresh(self):
        """
        :obj:`ipywidgets.FloatText`
        """
        return self._y_thresh

    @property
    def y_min(self):
        """
        :obj:`ipywidgets.Text`
        """
        return self._y_min

    @property
    def y_max(self):
        """
        :obj:`ipywidgets.Text`
        """
        return self._y_max

    @property
    def z(self):
        """
        :obj:`ipywidgets.Dropdown`
        """
        return self._z

    @property
    def z_active(self):
        """
        :obj:`ipywidgets.Checkbox`
        """
        return self._z_active

    @property
    def z_log(self):
        """
        :obj:`ipywidgets.Checkbox`
        """
        return self._z_log

    @property
    def z_thresh(self):
        """
        :obj:`ipywidgets.FloatText`
        """
        return self._z_thresh

    @property
    def z_min(self):
        """
        :obj:`ipywidgets.Text`
        """
        return self._z_min

    @property
    def z_max(self):
        """
        :obj:`ipywidgets.Text`
        """
        return self._z_max

    def axes_pannels_trigger(self):
        self.axes_options.children = [
            self.axes_pannels,
            self.panels[self.axes_pannels.value],
        ]

    def get_channel(self, channel):
        obj, _ = self.get_selected_entities()

        if channel is None:
            return None

        if channel not in self.data_channels.keys():

            if obj.get_data(channel):
                values = np.asarray(obj.get_data(channel)[0].values, dtype=float).copy()
                values[(values > 1e-38) * (values < 2e-38)] = np.nan
            elif channel == "Z":
                # Check number of points
                if hasattr(obj, "centroids"):
                    values = obj.centroids[:, 2]
                elif hasattr(obj, "vertices"):
                    values = obj.vertices[:, 2]
            else:
                return

            self.data_channels[channel] = values

        return self.data_channels[channel].copy()

    def set_channel_bounds(self, name):
        """
        Set the min and max values for the given axis channel
        """

        channel = getattr(self, "_" + name).value
        self.get_channel(channel)

        if channel in self.data_channels.keys():

            values = self.data_channels[channel]
            values = values[~np.isnan(values)]

            cmin = getattr(self, "_" + name + "_min")
            cmin.value = f"{np.min(values):.2e}"
            cmax = getattr(self, "_" + name + "_max")
            cmax.value = f"{np.max(values):.2e}"

    def plot_selection(
        self,
        x,
        x_log,
        x_active,
        x_thresh,
        x_min,
        x_max,
        y,
        y_log,
        y_active,
        y_thresh,
        y_min,
        y_max,
        z,
        z_log,
        z_active,
        z_thresh,
        z_min,
        z_max,
        color,
        color_log,
        color_active,
        color_thresh,
        color_maps,
        color_min,
        color_max,
        size,
        size_log,
        size_active,
        size_thresh,
        size_markers,
        size_min,
        size_max,
        refresh_trigger,
    ):

        if (
            not self.refresh_trigger.value
            or not self.refresh.value
            or self.indices is None
        ):
            return None

        if (
            self.downsampling.value != self.n_values
            and self.indices.shape[0] == self.n_values
        ):
            return self.update_downsampling(None)

        if self.get_channel(size) is not None and size_active:
            vals = self.get_channel(size)[self.indices]
            inbound = (vals > size_min) * (vals < size_max)
            vals[~inbound] = np.nan
            size = normalize(vals)

            if size_log:
                size = symlog(size, size_thresh)
            size *= size_markers
        else:
            size = None

        if self.get_channel(color) is not None and color_active:
            vals = self.get_channel(color)[self.indices]
            inbound = (vals >= color_min) * (vals <= color_max)
            vals[~inbound] = np.nan
            color = normalize(vals)
            if color_log:
                color = symlog(color, color_thresh)
        else:
            color = "black"

        x_axis, y_axis, z_axis = None, None, None

        if np.sum([x_active, y_active, z_active]) > 1:

            if x_active:
                x_axis = self.get_channel(x)
                if x_axis is None:
                    x_active = False
                else:
                    x_axis = x_axis[self.indices]

            if y_active:
                y_axis = self.get_channel(y)
                if y_axis is None:
                    y_active = False
                else:
                    y_axis = y_axis[self.indices]

            if z_active:
                z_axis = self.get_channel(z)
                if z_axis is None:
                    z_active = False
                else:
                    z_axis = z_axis[self.indices]

            if np.sum([axis is not None for axis in [x_axis, y_axis, z_axis]]) < 2:
                self.figure.data = []
                return

            if x_axis is not None:
                inbound = (x_axis >= x_min) * (x_axis <= x_max)
                x_axis[~inbound] = np.nan
                x_axis, x_label, x_ticks, x_ticklabels = format_axis(
                    x, x_axis, x_log, x_thresh
                )
            else:
                inbound = (z_axis >= z_min) * (z_axis <= z_max)
                z_axis[~inbound] = np.nan
                x_axis, x_label, x_ticks, x_ticklabels = format_axis(
                    z, z_axis, z_log, z_thresh
                )

            if y_axis is not None:
                inbound = (y_axis >= y_min) * (y_axis <= y_max)
                y_axis[~inbound] = np.nan
                y_axis, y_label, y_ticks, y_ticklabels = format_axis(
                    y, y_axis, y_log, y_thresh
                )
            else:
                inbound = (z_axis >= z_min) * (z_axis <= z_max)
                z_axis[~inbound] = np.nan
                y_axis, y_label, y_ticks, y_ticklabels = format_axis(
                    z, z_axis, z_log, z_thresh
                )

            if z_axis is not None:
                inbound = (z_axis >= z_min) * (z_axis <= z_max)
                z_axis[~inbound] = np.nan
                z_axis, z_label, z_ticks, z_ticklabels = format_axis(
                    z, z_axis, z_log, z_thresh
                )

            if self.custom_colormap:
                color_maps = self.custom_colormap

            # 3D Scatter
            if np.sum([x_active, y_active, z_active]) == 3:

                plot = go.Scatter3d(
                    x=x_axis,
                    y=y_axis,
                    z=z_axis,
                    mode="markers",
                    marker={"color": color, "size": size, "colorscale": color_maps},
                )

                layout = {
                    "margin": dict(l=0, r=0, b=0, t=0),
                    "scene": {
                        "xaxis_title": x_label,
                        "yaxis_title": y_label,
                        "zaxis_title": z_label,
                        "xaxis": {
                            "tickvals": x_ticks,
                            # "ticktext": [f"{label:.2e}" for label in x_ticklabels],
                        },
                        "yaxis": {
                            "tickvals": y_ticks,
                            # "ticktext": [f"{label:.2e}" for label in y_ticklabels],
                        },
                        "zaxis": {
                            "tickvals": z_ticks,
                            # "ticktext": [f"{label:.2e}" for label in z_ticklabels],
                        },
                    },
                }
            # 2D Scatter
            else:
                plot = go.Scatter(
                    x=x_axis,
                    y=y_axis,
                    mode="markers",
                    marker={"color": color, "size": size, "colorscale": color_maps},
                )

                layout = {
                    "margin": dict(l=0, r=0, b=0, t=0),
                    "xaxis": {
                        "tickvals": x_ticks,
                        # "ticktext": [f"{label:.2e}" for label in x_ticklabels],
                        "exponentformat": "e",
                        "title": x_label,
                    },
                    "yaxis": {
                        "tickvals": y_ticks,
                        # "ticktext": [f"{label:.2e}" for label in y_ticklabels],
                        "exponentformat": "e",
                        "title": y_label,
                    },
                }

            self.figure.data = []
            self.figure.add_trace(plot)
            self.figure.update_layout(layout)

        else:
            self.figure.data = []

    def update_axes(self, refresh_plot=True):
        self.refresh_trigger.value = False
        for name in [
            "x",
            "y",
            "z",
            "color",
            "size",
        ]:
            widget = getattr(self, "_" + name)
            val = widget.value
            widget.options = list(self.data_channels.keys())

            if val in widget.options:
                widget.value = val
            else:
                widget.value = None
        if refresh_plot:
            self.refresh_trigger.value = True

    def update_choices(self, _):
        self.refresh_trigger.value = False

        for channel in self.data.value:
            self.get_channel(channel)

        keys = list(self.data_channels.keys())
        for key in keys:
            if key not in self.data.value:
                del self.data_channels[key]

        self.update_axes(refresh_plot=False)

        if self.downsampling.value != self.n_values:
            self.update_downsampling(None, refresh_plot=False)

        self.refresh_trigger.value = True

    def update_downsampling(self, _, refresh_plot=True):

        if not list(self.data_channels.values()):
            return

        self.refresh_trigger.value = False
        values = []
        for axis in [self.x, self.y, self.z]:
            vals = self.get_channel(axis.value)
            if vals is not None:
                values.append(np.asarray(vals, dtype=float))

        if len(values) < 2:
            return

        values = np.vstack(values)
        nans = np.isnan(values)
        values[nans] = 0
        # Normalize all columns
        values = (values - np.min(values, axis=1)[:, None]) / (
            np.max(values, axis=1) - np.min(values, axis=1)
        )[:, None]
        values[nans] = np.nan
        self._indices = random_sampling(
            values.T,
            self.downsampling.value,
            bandwidth=2.0,
            rtol=1e0,
            method="histogram",
        )
        self.refresh_trigger.value = refresh_plot

    def update_objects(self, _):
        self.data_channels = {}
        self.downsampling.max = self.n_values
        self.downsampling.value = np.min([5000, self.n_values])
        self._indices = None
        self.update_downsampling(None, refresh_plot=False)

    def write_html(self):
        self.figure.write_html(
            os.path.join(
                os.path.abspath(os.path.dirname(self.h5file)), "Crossplot.html"
            )
        )



#  Copyright (c) 2021 Mira Geoscience Ltd.
#
#  This file is part of geoapps.
#
#  geoapps is distributed under the terms and conditions of the MIT License
#  (see LICENSE file at the root of this source code package).

from .calculator import Calculator
from .clustering import Clustering
from .coordinate_transformation import CoordinateTransformation
from .create import ContourValues, IsoSurface, Surface2D
from .data_interpolation import DataInterpolation
from .edge_detection import EdgeDetectionApp
from .peak_finder import PeakFinder



#  Copyright (c) 2021 Mira Geoscience Ltd.
#
#  This file is part of geoapps.
#
#  geoapps is distributed under the terms and conditions of the MIT License
#  (see LICENSE file at the root of this source code package).

import numpy as np
from geoh5py.io import H5Writer
from geoh5py.objects import Curve, Grid2D
from geoh5py.workspace import Workspace
from ipywidgets import (
    FloatSlider,
    HBox,
    IntSlider,
    Layout,
    Text,
    ToggleButton,
    VBox,
    interactive_output,
)
from matplotlib import collections
from skimage.feature import canny
from skimage.transform import probabilistic_hough_line

from geoapps.plotting import PlotSelection2D
from geoapps.utils import filter_xy


class EdgeDetectionApp(PlotSelection2D):
    """
    Widget for Grid2D objects for the automated detection of line features.
    The application relies on the Canny and Hough trandforms from the
    Scikit-Image library.

    :param grid: Grid2D object
    :param data: Children data object for the provided grid

    Optional
    --------

    :param sigma [Canny]: standard deviation of the Gaussian filter
    :param threshold [Hough]: Value threshold
    :param line_length [Hough]: Minimum accepted pixel length of detected lines
    :param line_gap [Hough]: Maximum gap between pixels to still form a line.
    """

    defaults = {
        "h5file": "../../assets/FlinFlon.geoh5",
        "objects": "Gravity_Magnetics_drape60m",
        "data": "Airborne_Gxx",
        "resolution": 50,
        "sigma": 0.5,
        "compute": True,
        "window": {
            "azimuth": -20,
        },
        "ga_group_name": "Edges",
    }
    object_types = (Grid2D,)

    def __init__(self, **kwargs):

        self._compute = ToggleButton(
            value=False,
            description="Compute",
            button_style="warning",
            tooltip="Description",
            icon="check",
        )
        self._export_as = Text(
            value="Edges",
            description="Save as:",
            disabled=False,
        )
        self._line_length = IntSlider(
            min=1,
            max=100,
            step=1,
            value=1,
            continuous_update=False,
            description="Line Length",
        )
        self._line_gap = IntSlider(
            min=1,
            max=100,
            step=1,
            value=1,
            continuous_update=False,
            description="Line Gap",
        )
        self._sigma = FloatSlider(
            min=0.0,
            max=10,
            step=0.1,
            value=1.0,
            continuous_update=False,
            description="Sigma",
        )
        self._threshold = IntSlider(
            min=1,
            max=100,
            step=1,
            value=1,
            continuous_update=False,
            description="Threshold",
        )
        self._window_size = IntSlider(
            min=16,
            max=512,
            value=64,
            continuous_update=False,
            description="Window size",
        )
        super().__init__(**kwargs)

        out = interactive_output(
            self.compute_trigger,
            {"compute": self.compute},
        )

        def save_trigger(_):
            self.save_trigger()

        # Make changes to trigger warning color
        self.trigger.description = "Save to GA"
        self.trigger.on_click(save_trigger)
        self.trigger.button_style = "success"

        def update_name(_):
            self.update_name()

        self.data.observe(update_name, names="value")
        self.update_name()

        self._main = VBox(
            [
                self.project_panel,
                HBox(
                    [
                        self.main,
                        VBox(
                            [
                                self.sigma,
                                self.threshold,
                                self.line_length,
                                self.line_gap,
                                self.window_size,
                                self.compute,
                                self.export_as,
                                self.output_panel,
                            ],
                            layout=Layout(width="50%"),
                        ),
                        out,
                    ]
                ),
            ]
        )

    @property
    def compute(self):
        """ToggleButton"""
        return self._compute

    @property
    def export_as(self):
        """Text"""
        return self._export_as

    @property
    def line_length(self):
        """IntSlider"""
        return self._line_length

    @property
    def line_gap(self):
        """IntSlider"""
        return self._line_gap

    @property
    def sigma(self):
        """FloatSlider"""
        return self._sigma

    @property
    def threshold(self):
        """IntSlider"""
        return self._threshold

    @property
    def window_size(self):
        """IntSlider"""
        return self._window_size

    def save_trigger(self):
        entity, _ = self.get_selected_entities()
        if getattr(self.trigger, "vertices", None) is not None:

            curves = [
                child
                for child in self.ga_group.children
                if child.name == self.export_as.value
            ]
            if any(curves):
                curve = curves[0]

                curve._children = []
                curve.vertices = self.trigger.vertices
                curve.cells = np.vstack(self.trigger.cells).astype("uint32")

                # Remove directly on geoh5
                project_handle = H5Writer.fetch_h5_handle(self.h5file, entity)
                base = list(project_handle.keys())[0]
                obj_handle = project_handle[base]["Objects"]
                for key in obj_handle[H5Writer.uuid_str(curve.uid)]["Data"].keys():
                    del project_handle[base]["Data"][key]
                del obj_handle[H5Writer.uuid_str(curve.uid)]

                H5Writer.save_entity(curve)

            else:
                curve = Curve.create(
                    self.workspace,
                    name=self.export_as.value,
                    vertices=self.trigger.vertices,
                    cells=self.trigger.cells,
                    parent=self.ga_group,
                )

            if self.live_link.value:
                self.live_link_output(self.ga_group)

            self.workspace.finalize()

    def update_name(self):
        if self.data.value is not None:
            self.export_as.value = self.data.value
        else:
            self.export_as.value = "Edges"

    def compute_trigger(self, compute):
        if compute:

            grid, data = self.get_selected_entities()

            x = grid.centroids[:, 0].reshape(grid.shape, order="F")
            y = grid.centroids[:, 1].reshape(grid.shape, order="F")
            z = grid.centroids[:, 2].reshape(grid.shape, order="F")
            grid_data = data[0].values.reshape(grid.shape, order="F")
            indices = self.indices
            ind_x, ind_y = (
                np.any(indices, axis=1),
                np.any(indices, axis=0),
            )
            x = x[ind_x, :][:, ind_y]
            y = y[ind_x, :][:, ind_y]
            z = z[ind_x, :][:, ind_y]
            grid_data = grid_data[ind_x, :][:, ind_y]
            grid_data -= grid_data.min()
            grid_data /= grid_data.max()

            if np.any(grid_data):
                # Find edges
                edges = canny(grid_data, sigma=self.sigma.value, use_quantiles=True)
                shape = edges.shape
                # Cycle through tiles of square size
                max_l = np.min([self.window_size.value, shape[0], shape[1]])
                half = np.floor(max_l / 2)
                overlap = 1.25

                n_cell_y = (shape[0] - 2 * half) * overlap / max_l
                n_cell_x = (shape[1] - 2 * half) * overlap / max_l

                if n_cell_x > 0:
                    cnt_x = np.linspace(
                        half, shape[1] - half, 2 + int(np.round(n_cell_x)), dtype=int
                    ).tolist()
                    half_x = half
                else:
                    cnt_x = [np.ceil(shape[1] / 2)]
                    half_x = np.ceil(shape[1] / 2)

                if n_cell_y > 0:
                    cnt_y = np.linspace(
                        half, shape[0] - half, 2 + int(np.round(n_cell_y)), dtype=int
                    ).tolist()
                    half_y = half
                else:
                    cnt_y = [np.ceil(shape[0] / 2)]
                    half_y = np.ceil(shape[0] / 2)

                coords = []
                for cx in cnt_x:
                    for cy in cnt_y:

                        i_min, i_max = int(cy - half_y), int(cy + half_y)
                        j_min, j_max = int(cx - half_x), int(cx + half_x)
                        lines = probabilistic_hough_line(
                            edges[i_min:i_max, j_min:j_max],
                            line_length=self.line_length.value,
                            threshold=self.threshold.value,
                            line_gap=self.line_gap.value,
                            seed=0,
                        )

                        if np.any(lines):
                            coord = np.vstack(lines)
                            coords.append(
                                np.c_[
                                    x[i_min:i_max, j_min:j_max][
                                        coord[:, 1], coord[:, 0]
                                    ],
                                    y[i_min:i_max, j_min:j_max][
                                        coord[:, 1], coord[:, 0]
                                    ],
                                    z[i_min:i_max, j_min:j_max][
                                        coord[:, 1], coord[:, 0]
                                    ],
                                ]
                            )
                if coords:
                    coord = np.vstack(coords)
                    self.objects.lines = coord
                    self.plot_store_lines()
                else:
                    self.objects.lines = None

            self.compute.value = False

    def plot_store_lines(self):

        xy = self.objects.lines
        indices_1 = filter_xy(
            xy[1::2, 0],
            xy[1::2, 1],
            self.resolution.value,
            window={
                "center": [
                    self.center_x.value,
                    self.center_y.value,
                ],
                "size": [
                    self.width.value,
                    self.height.value,
                ],
                "azimuth": self.azimuth.value,
            },
        )
        indices_2 = filter_xy(
            xy[::2, 0],
            xy[::2, 1],
            self.resolution.value,
            window={
                "center": [
                    self.center_x.value,
                    self.center_y.value,
                ],
                "size": [
                    self.width.value,
                    self.height.value,
                ],
                "azimuth": self.azimuth.value,
            },
        )

        indices = np.kron(
            np.any(np.c_[indices_1, indices_2], axis=1),
            np.ones(2),
        ).astype(bool)

        xy = self.objects.lines[indices, :2]
        self.collections = [
            collections.LineCollection(
                np.reshape(xy, (-1, 2, 2)), colors="k", linewidths=2
            )
        ]
        self.refresh.value = False
        self.refresh.value = True  # Trigger refresh

        if np.any(xy):
            vertices = np.vstack(self.objects.lines[indices, :])
            cells = np.arange(vertices.shape[0]).astype("uint32").reshape((-1, 2))
            if np.any(cells):
                self.trigger.vertices = vertices
                self.trigger.cells = cells
        else:
            self.trigger.vertices = None
            self.trigger.cells = None
